#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/07 17:47:22
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''

from typing import List
import requests
from xagents.model.common import LLM, EMBD, Reranker
# from agit.backend.zhipuai_bk import call_llm_api
from agit.llm import call_llm, LLMResp
from agit.common import ToolCall
from agit.embd import call_embedding, EMBD_TYPE

from xagents.tool.core import ToolCall
from xagents.tool import BaseTool

from loguru import logger


class API_GLM(LLM):
    def __init__(self, version: str, name: str = None,  api_key=None):
        assert version in self.list_versions(), f"{version} not in {self.list_versions()}"
        super().__init__(name=name, version=version)
        self.api_key = api_key

    @classmethod
    def list_versions(cls):
        return [
            "glm-4",
            "glm-3-turbo",
            "chatglm3_130b",
            "chatglm_turbo",
            "chatglm_pro",
            "chatglm_66b",
            "chatglm_12b",
            "chatglm2_12b_32k"
        ]

    def observe(self, prompt: str, tool_call: ToolCall,  tools: List[BaseTool] = [], history=[], stream=True, **kwargs):
        observe_message = dict(role="tool", content=str(tool_call.resp), tool_call_id=tool_call.tool_call_id)
        messages = history + [dict(role="user", content=prompt)] + [observe_message]

        resp = call_llm(messages=messages, tools=tools,  model=self.version, do_search=False, stream=stream, api_key=self.api_key, **kwargs)
        logger.debug(f"observe result:{resp}")
        return resp

    def generate(self, prompt, history=[], system=None, tools: List[BaseTool] = [], stream=True, **kwargs) -> LLMResp:
        # logger.info(f"{self.__class__} generating resp with {prompt=}, {history=}")
        messages = history + [dict(role="user", content=prompt)]
        resp: LLMResp = call_llm(messages=messages, model=self.version, tools=tools, do_search=False,
                                 system=system, stream=stream, api_key=self.api_key, **kwargs)
        return resp


class ZhipuEmbedding(EMBD):
    def __init__(self,  api_key=None, batch_size=16, norm=True):
        self.api_key = api_key
        self.batch_size = batch_size
        self.norm = norm

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        embeddings = call_embedding(text=texts, embd_type=EMBD_TYPE.ZHIPU_API, model_or_url="embedding-2",
                                    norm=self.norm, batch_size=self.batch_size)
        return embeddings

    def embed_query(self, text: str) -> List[float]:
        embedding = call_embedding(text=text, embd_type=EMBD_TYPE.ZHIPU_API, model_or_url="embedding-2",
                                   norm=self.norm, batch_size=self.batch_size)
        return embedding

    def get_dim(self) -> int:
        return 1024


class ZhipuReranker(Reranker):
    def __init__(self,  url: str, name="reranker", version="ZhipuReranker"):
        self.url = url
        super().__init__(name=name, version=version)

    def cal_similarity(self, text1: str, text2: str):
        logger.debug(f"rerank simi for {text1}, {text2}")
        resp = requests.post(url=self.url, params=dict(text1=text1, text2=text2))
        resp.raise_for_status()
        return resp.json()["data"]["score"]


if __name__ == "__main__":
    # llm_model = GLM(name="glm", version="chatglm_turbo")
    # resp = llm_model.generate("你好", stream=False)
    # print(resp)

    # embd_model = ZhipuEmbedding()
    # text = ["中国", "美国", "日本", "法国", "英国", "意大利", "西班牙", "德国", "俄罗斯"]
    # embds = embd_model.embed_documents(text)
    # print(len(embds))
    # print(embds[0][:4])
    # embd = embd_model.embed_query("你好")
    # print(len(embd))
    # print(embd[:4])

    reranker = ZhipuReranker(url="http://hz-model.bigmodel.cn/reranker/get_rel_score", name="zhipu_ranker", version="bge-reranker-base")
    sim = reranker.cal_similarity("私募基金", "公募基金")
    print(sim)



import re
from fake_useragent import UserAgent
from agit.common import Parameter
from xagents.tool.core import BaseTool
import requests
from bs4 import BeautifulSoup
import time
import random

import requests
import chardet


def get_search_result(url):
    headers = {
        "User-Agent": UserAgent().random,  # Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36"
    }
    try:
        response = requests.get(url, headers=headers)
        # 使用 chardet 检测网页编码
        encoding = chardet.detect(response.content)['encoding']
        # 设置解码方式
        response.encoding = encoding

        if response.status_code == 200:
            soup = BeautifulSoup(response.text, "html.parser")
            # 解析搜索结果页面，这里以输出标题和链接为例
            soup_text = soup.text
            return soup_text
    except:
        return ""


def sogou_search(query):
    url = f"https://www.sogou.com/web"
    # "http://www.so.com/s?q=keyword"
    # f"https://www.sogou.com/web"
    params = {"query": query}

    headers = {
        "User-Agent": UserAgent().random,
        # "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36"
    }

    try:
        response = requests.get(url, params=params, headers=headers)

        # 使用 chardet 检测网页编码
        encoding = chardet.detect(response.content)['encoding']
        # 设置解码方式
        response.encoding = encoding

        # 等待一段随机时间，模拟人类操作
        time.sleep(random.uniform(1, 2))
        sub_page_contents = []
        sub_page_links = []
        if response.status_code == 200:
            soup = BeautifulSoup(response.text, "html.parser")
            # 解析搜索结果页面，这里以输出标题和链接为例
            soup_text = soup.text  # str(soup)
            pattern = r'data-url="(.*?)"'
            urls = re.findall(pattern, soup_text)
            if len(urls) >= 1:
                for url_temp in urls:
                    sub_page_links.append(url_temp)
                    sub_page_contents.append(get_search_result(url_temp))
            else:
                sub_page_links = []
                sub_page_contents.append(soup_text.replace("\n\n", ""))

            return sub_page_links, sub_page_contents
        else:
            print(f"Error: Failed to fetch search results (status code {response.status_code})")
            return None, None
    except Exception as e:
        print(f"Error: {str(e)}")
        return None, None


#
# def baidu_search(keyword):
#     url = "https://www.baidu.com/s"
#     params = {
#         "wd": keyword  # 关键词参数
#     }
#
#     headers = {
#         "User-Agent": UserAgent().random,#"Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36",
#         "Referer": "https://www.baidu.com/",  # 设置 Referer 头部
#         "Accept-Language": "en-US,en;q=0.9",  # 接受的语言
#     }
#
#     try:
#         # 发送 GET 请求
#         response = requests.get(url, params=params, headers=headers)
#         response.raise_for_status()  # 抛出异常，如果响应码不是 200
#
#         # 使用 chardet 检测网页编码
#         encoding = chardet.detect(response.content)['encoding']
#         # 设置解码方式
#         response.encoding = encoding
#
#         # 等待一段随机时间，模拟人类操作
#         time.sleep(random.uniform(1, 3))
#
#         # 解析 HTML 页面
#         soup = BeautifulSoup(response.text, "html.parser")
#
#         # 提取搜索结果
#         results = []
#         sub_page_contents = []
#         for item in soup.find_all("h3", class_="t"):
#             title = item.get_text()
#             link = item.a["href"]
#             results.append({"title": title, "link": link})
#             sub_page_content = open_web_page(link)
#             sub_page_contents.append(sub_page_content)
#
#         return results, sub_page_contents
#
#     except requests.exceptions.HTTPError as e:
#         print(f"HTTP Error occurred: {e}")
#     except requests.exceptions.RequestException as e:
#         print(f"An error occurred: {e}")
#
#     return None,None


def baidu_search(query):
    url = "https://www.baidu.com/s"
    params = {
        "wd": query  # 关键词参数
    }

    headers = {
        "User-Agent": UserAgent().random,
        # "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36",
        "Referer": "https://www.baidu.com/",  # 设置 Referer 头部
        "Accept-Language": "en-US,en;q=0.9",  # 接受的语言
    }

    try:
        # 发送 GET 请求
        response = requests.get(url, params=params, headers=headers)
        response.raise_for_status()  # 抛出异常，如果响应码不是 200
        # 使用 chardet 检测网页编码
        encoding = chardet.detect(response.content)['encoding']
        # 设置解码方式
        response.encoding = encoding
        # 等待一段随机时间，模拟人类操作
        time.sleep(random.uniform(1, 3))

        # 解析 HTML 页面
        soup = BeautifulSoup(response.text, "html.parser")
        soup_text = str(soup)
        pattern = r'"mu":"(.*?)"'
        urls = re.findall(pattern, soup_text)
        # 提取搜索结果
        results = []
        sub_page_contents = []
        for url in urls:
            sub_page_contents.append(url)

        return sub_page_contents

    except requests.exceptions.HTTPError as e:
        print(f"HTTP Error occurred: {e}")
    except requests.exceptions.RequestException as e:
        print(f"An error occurred: {e}")

    return None

# print(fetch_baidu_search_results("江泽民生平"))


web_search_sogou = BaseTool(name="联网查询", description="根据用户提供的输入，通过搜索引擎从网上去查询信息，尤其是实时的信息",
                            parameters=[Parameter(name="query", description="搜索的语句", type="string", required=True)],
                            callable=sogou_search)


# 测试搜索功能
if __name__ == "__main__":
    keyword = "邓小平生平"
    # web_search = WebSearch()
    # search_results = web_search.sogou_search(keyword)
    search_links, search_contents = web_search_sogou.execute(query=keyword)
    print(f"Search results for '{search_links}':")
    print("text:", search_contents)
    # from snippets.logs import ChangeLogLevelContext



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/01/04 14:01:01
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''

from typing import Any, Callable

from pydantic import Field
from agit.common import ToolDesc, ToolCall, ToolDesc



class BaseTool(ToolDesc):
    callable:Callable = Field(..., description="工具执行函数")
    
    
    def execute(self, *args, **kwargs) -> Any:
        return self.callable(*args, **kwargs)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/07 18:40:00
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''
import os

XAGENT_ENV = os.environ.get("XAGENT_ENV", "dev")
XAGENT_HOME = os.path.dirname(os.path.dirname(__file__))

KNOWLEDGE_BASE_DIR = os.environ.get("XAGENT_KNOWLEDGE_BASE_DIR", os.path.join(XAGENT_HOME, "knowledge_base"))
os.makedirs(KNOWLEDGE_BASE_DIR, exist_ok=True)

TEMP_DIR = os.environ.get("XAGENT_TEMP_DIR", os.path.join(XAGENT_HOME, "tmp"))
os.makedirs(TEMP_DIR, exist_ok=True)

LOG_DIR = os.environ.get("XAGENT_LOG_DIR", os.path.join(XAGENT_HOME, "log"))
os.makedirs(LOG_DIR, exist_ok=True)

JOB_LOG_DIR = os.path.join(LOG_DIR, "jobs")
os.makedirs(JOB_LOG_DIR, exist_ok=True)

DB_DIR = os.environ.get("XAGENT_DB_DIR", os.path.join(XAGENT_HOME, "db"))
os.makedirs(DB_DIR, exist_ok=True)

AGENT_DIR = os.environ.get("XAGENT_AGENT_DIR", os.path.join(XAGENT_HOME, "agent"))
os.makedirs(AGENT_DIR, exist_ok=True)

DATA_DIR = os.path.join(XAGENT_HOME, "data")


DEFAULT_KB_PROMPT_TEMPLATE = '''请根据[参考信息]回答我的问题，如果问题不在参考信息内，请不要参考
[参考信息]
{context}

问题:
{question}
'''

DEFAULT_WEB_SEARCH_PROMPT_TEMPLATE = '''请根据[搜索信息]回答我的问题，如果问题不在搜索信息内，请不要参考
[搜索信息]
{search_info}
问题:
{question}
'''

DEFAULT_WEB_SEARCH_KB_PROMPT_TEMPLATE = '''请根据[参考信息]和[搜索信息]回答我的问题，如果问题不在参考信息和搜索信息内，请不要参考
[参考信息]
{context}
[搜索信息]
{search_info}
问题:
{question}
'''


# Service相关
# 最大job数（job用来处理知识库index任务）
MAX_JOB_NUM = 10
# 用户鉴权配置
USERNAME = "zhipu"
PASSWORD = "zhipu"
# xagent缓存数目
XAGENT_CACHE_NUM = 1000
XAGENT_CACHE_EXPIRE_SECONDS = 3600


# MinIO对象存储相关
MINIO_URL = os.environ.get("MINIO_URL", "10.50.130.151:9000")
MINIO_ACCESS_KEY = os.environ.get("MINIO_ACCESS_KEY", 'l7RJ3QCGX6gt7M8zfN1v')
MINIO_SECRET_KEY = os.environ.get("MINIO_SECRET_KEY", 'smGpFhcz0hLOup8V2s6SMkzgzxzJFtSJiuAuKFqS')



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

from typing import Iterable



from xagents.config import *
from xagents.loader.common import Chunk, AbstractLoader, TableChunk, data2markdown, merge_chunks

import pandas as pd

def analyse_table(file_path, **kwargs)->Iterable[Chunk]:

    xl = pd.ExcelFile(file_path)
    for name in xl.sheet_names:
        df:pd.DataFrame = pd.read_excel(xl, name)

        page_index = 1
        
        # content = df.to_json()
        content = data2markdown(df)
        yield TableChunk(content=content,page_idx=page_index, data=df)
        page_index+=1

            
class EXCELLoader(AbstractLoader):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


    def load(self, file_path: str, **kwargs) -> Iterable[Chunk]:
        #TODO 缺少页码控制
        chunks = analyse_table(file_path, **kwargs)
        chunks = merge_chunks(chunks)
        yield from chunks


if __name__ == "__main__":
    chunks = analyse_table("data/kb_file/大学综合排名2022.xlsx")
    for chunk in chunks:
        print(chunk)


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/08 10:14:40
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''

from agit.common import *
from xagents.agent.xagent import XAgent


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/12 11:34:29
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''

from typing import List
from fastapi import UploadFile
from loguru import logger
from pydantic import BaseModel, Field
from xagents.loader.common import Chunk
from xagents.kb.kb_file import KnowledgeBaseFile
from xagents.kb.kb import KnowledgeBase
from xagents.config import *
from xagents.kb.common import KnowledgeBaseInfo, KnowledgeBaseFileInfo, RecalledChunk, get_config_path, DistanceStrategy


def list_knowledge_base_names() -> List[str]:
    """列出所有知识库名称

    Returns:
        str: 知识库名称列表
    """
    kb_names = os.listdir(KNOWLEDGE_BASE_DIR)
    return kb_names


def list_knowledge_base_info() -> List[KnowledgeBaseInfo]:
    kb_infos = []
    kb_names = list_knowledge_base_names()
    for name in kb_names:
        kb = get_knowledge_base(name)
        kb_infos.append(kb.get_info())
    return kb_infos


def get_knowledge_base(name: str) -> KnowledgeBase:
    """根据知识库名称获取知识库

    Args:
        name (str): 知识库名称

    Raises:
        ValueError: 知识库不存在异常

    Returns:
        KnowledgeBase: 知识库实例
    """
    config_path = get_config_path(name)
    if not os.path.exists(config_path):
        message = f"知识库配置文件不存在: {config_path}"
        logger.error(message)
        raise ValueError(message)
    kb = KnowledgeBase.from_config(config_path)
    return kb


def get_knowledge_base_info(name: str) -> KnowledgeBaseInfo:
    """
    根据名称获取知识库实例
    """
    kb = get_knowledge_base(name=name)
    return kb.get_info()


def create_knowledge_base(name: str,
                          desc: str = None,
                          embedding_config: dict = dict(cls="ZhipuEmbedding"),
                          vecstore_config: dict | List[dict] = dict(cls='XFAISS', distance_strategy=DistanceStrategy.MAX_INNER_PRODUCT)
                          ) -> KnowledgeBase:
    """创建知识库

    Raises:
        ValueError: 知识库已经存在的异常

    Returns:
        _type_: _description_
    """

    logger.info(f"Creating knowledge base {name}...")
    kb_names = list_knowledge_base_names()
    if name in kb_names:
        msg = f"Knowledge base {name} already exists, can not create!"
        logger.warning(msg)
        raise ValueError(msg)
    if not desc:
        desc = f"知识库{name}"

    kb = KnowledgeBase(name=name, embedding_config=embedding_config,
                       desc=desc, vecstore_config=vecstore_config)
    return kb


def delete_knowledge_base(name: str) -> str:
    """删除知识库

    Args:
        name (str): 知识库名称

    Returns:
        str: 删除消息
    """
    # delete knowledge base
    kb = get_knowledge_base(name=name)
    kb.delete()

    # kb_dir = get_kb_dir(kb_name=name)
    # if os.path.exists(kb_dir):
    #     import shutil
    #     shutil.rmtree(path=kb_dir)
    msg = f'【{name}】deleted.'
    return msg


def reindex_knowledge_base(name: str, batch_size=16) -> str:
    """重新构建知识库索引

    Args:
        name (str): 知识库名称
        batch_size(int): 调用embedding的时候的batch_size（对于开发平台api,是并发度）
    """
    kb = get_knowledge_base(name=name)
    kb.rebuild_index(batch_size=batch_size)
    msg = f"知识库【{name}】重建索引成功"
    return msg


class KBSearchConfig(BaseModel):
    top_k: int = Field(default=3, description="返回几个chunk")
    score_threshold: float = Field(default=0., description="分数阈值")
    do_split_query: bool = Field(default=False, description="是否分词")
    file_names: List[str] = Field(default=[], description="需要筛选的知识库文件文件名列表")
    rerank_config: dict = Field(default={}, description="rerank的配置")
    do_expand: bool = Field(default=False, description="对结果是否进行上下文扩展")
    expand_len: int = Field(default=500, description="上下文扩展的长度")
    forward_rate: float = Field(default=0.5, description="上下文扩展向下扩展的比例")


def search_knowledge_base(name: str,
                          query: str, top_k: int = 3, score_threshold: float = None,
                          do_split_query=False, file_names: List[str] = [], rerank_config: dict = {},
                          do_expand=False, expand_len: int = 500, forward_rate: float = 0.5) -> List[RecalledChunk]:
    kb = get_knowledge_base(name=name)

    chunks = kb.search(query=query, top_k=top_k, score_threshold=score_threshold, do_split_query=do_split_query,
                       file_names=file_names, rerank_config=rerank_config, do_expand=do_expand, expand_len=expand_len, forward_rate=forward_rate)
    return chunks


def list_kb_files(kb_name: str) -> List[KnowledgeBaseFile]:
    kb = get_knowledge_base(kb_name)
    kb_files = kb.list_kb_files()
    return kb_files


def list_kb_file_infos(kb_name: str) -> List[KnowledgeBaseFileInfo]:
    kb = get_knowledge_base(kb_name)
    kb_files = kb.list_kb_files()
    kb_file_infos = [kb_file.get_info() for kb_file in kb_files]
    return kb_file_infos


def get_kb_file(kb_name: str, file_name: str) -> KnowledgeBaseFile:
    kb_file = KnowledgeBaseFile(kb_name=kb_name, file_name=file_name)
    if os.path.exists(kb_file.origin_path):
        return kb_file
    raise ValueError(f"{kb_file.origin_path} not exists.")


def get_kb_file_info(kb_name: str, file_name: str) -> KnowledgeBaseFileInfo:
    kb_file = get_kb_file(kb_name=kb_name, file_name=file_name)
    return kb_file.get_info()


def create_kb_file(kb_name: str, file: UploadFile | str, do_cut=True, do_index=True,
                   batch_size: int = 16,
                   upload_image: bool = False,
                   ocr: bool = False,
                   cut_config: dict = dict(separator='\n',
                                           max_len=200,
                                           min_len=10)) -> KnowledgeBaseFile:
    """创建知识库文件
    Args:
        kb_name (str): 知识库名称
        file (UploadFile | str): 知识库文件，UploadFile(fast_api)或者str(文件路径)
        do_cut (bool, optional): 是否切分文件. Defaults to True.
        do_index (bool, optional): 是否添加到索引. Defaults to True.
        upload_image (bool, optional): 是否上传图片. Defaults to False.
        ocr (bool, optional): 是否OCR. Defaults to False.
        cut_config (dict, optional): 切分文件的参数. Defaults to dict(separator='\n', max_len=200, min_len=10).

    Raises:
        ValueError: 知识库文件已经存在

    Returns:
        KnowledgeBaseFileInfo: 知识库文件描述
    """
    if isinstance(file, str):
        file_name = os.path.basename(file)
    else:
        file_name = os.path.basename(file.filename)
    logger.debug(f"creating kb_file with {kb_name=}, {file_name=}")
    kb_file = KnowledgeBaseFile(kb_name=kb_name, file_name=file_name)
    kb_file_path = kb_file.origin_path
    if os.path.exists(kb_file_path):
        raise ValueError(f"{kb_file_path} already exists.")
    if isinstance(file, str):
        with open(file, "rb") as f:
            content = f.read()
    else:
        content = file.file.read()
    with open(kb_file_path, "wb") as f:
        f.write(content)
    if do_cut:
        kb_file.cut(upload_image=upload_image, ocr=ocr, **cut_config)
    if do_index:
        kb = get_knowledge_base(name=kb_name)
        kb.add_kb_file2index(kb_file=kb_file, do_save=True, batch_size=batch_size)
    return kb_file


def delete_kb_file(kb_name: str, file_name: str) -> str:
    """删除知识库文件

    Args:
        kb_name (str): 知识库名称
        file_name (str): 知识库文件名称

    Returns:
        str: 删除消息
    """

    kb = get_knowledge_base(kb_name)
    kb_file = get_kb_file(kb_name=kb_name, file_name=file_name)
    kb.remove_kb_file(kb_file=kb_file)
    return f"知识库文件【{file_name}】删除成功"


def cut_kb_file(kb_name: str, file_name: str,
                separator: str = '\n',
                max_len: int = 200,
                min_len: int = 10) -> dict:
    """切分文档，并且按照jsonl格式存储在chunk目录下

    Args:
        kb_name (str): 知识库名称
        file_name (str): 文件名称
        separator (str, optional): 切分符. Defaults to '\n'
        max_len (int, optional): 最大切片长度. Defaults to 200
        min_len (int, optional): 最小切片长度. Defaults to 10

    Returns:
        int: 切片数目
    """

    kb_file = get_kb_file(kb_name=kb_name, file_name=file_name)
    chunk_num = kb_file.cut(separator=separator, max_len=max_len, min_len=min_len)
    return dict(chunk_num=chunk_num)


def reindex_kb_file(kb_name: str, file_name: str, batch_size=16) -> str:
    """添加知识库文件到索引中

    Args:
        kb_name (str): 知识库名称
        file_name (str): 知识库文件名称
    Returns:
        str: 构建成功的消息
    """

    kb = get_knowledge_base(kb_name)
    kb_file = get_kb_file(kb_name=kb_name, file_name=file_name)
    kb.add_kb_file2index(kb_file=kb_file, do_save=True, batch_size=batch_size)
    return f"更新知识库文件【{file_name}】到索引成功"


def list_chunks(kb_name: str, file_name: str) -> List[Chunk]:
    """给定知识库文件，返回所有的chunk

    Args:
        kb_name (str): 知识库名称
        file_name (str): 知识库文件名称

    Returns:
        List[Chunk]: chunk列表
    """
    kb_file: KnowledgeBaseFile = get_kb_file(kb_name=kb_name, file_name=file_name)
    chunks = kb_file.list_chunks()
    return chunks


def add_chunks(kb_name: str, file_name: str, chunks: List[Chunk], idx: int = None) -> str:
    """给定知识库文件，添加chunk

    Args:
        kb_name (str): 知识库名称
        file_name (str): 知识库文件名称
        chunk (Chunk): 待添加的chunk
        idx (_type_, optional): chunk添加的位置，None的话添加在最后. Defaults to None.

    Returns:
        str: 添加成功的消息
    """
    kb_file: KnowledgeBaseFile = get_kb_file(kb_name=kb_name, file_name=file_name)
    kb_file.add_chunks(chunks=chunks, idx=idx)
    kb: KnowledgeBase = get_knowledge_base(name=kb_name)
    kb.add_chunks2index(chunks=chunks, meta_info=dict(file_name=file_name), do_save=True)
    message = f"添加{len(chunks)}chunk到知识库文件【{file_name}】成功"
    return message


def delete_chunks(kb_name: str, file_name: str, chunk_ids: List[str]) -> str:

    kb_file: KnowledgeBaseFile = get_kb_file(kb_name=kb_name, file_name=file_name)
    kb_file.delete_chunks(ids=chunk_ids)
    kb: KnowledgeBase = get_knowledge_base(name=kb_name)
    kb.remove_chunks_from_index(chunk_ids=chunk_ids)
    message = f"从【{file_name}】删除{len(chunk_ids)}chunk成功"
    return message


def update_chunk(kb_name: str, file_name: str, chunk: Chunk) -> str:
    kb_file: KnowledgeBaseFile = get_kb_file(kb_name=kb_name, file_name=file_name)
    kb_file.update_chunk(chunk=chunk)
    kb: KnowledgeBase = get_knowledge_base(name=kb_name)
    logger.debug(f"{chunk=}")
    kb.add_chunks2index(chunks=[chunk], meta_info=dict(file_name=file_name), do_save=True)


if __name__ == "__main__":
    # print(list_knowledge_base_names())
    # print(list_vecstores())
    # print(list_distance_strategy())
    # chunk_len = cut_kb_file(kb_name="new_kb", file_name="requirements.txt")
    # print(chunk_len)

    kb = get_knowledge_base(name="new_kb")
    chunks = kb.search(query="python-snippets", score_threshold=0.4)
    print(chunks)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/04/11 17:17:19
@Author  :   ChenHao
@Description  :   测试xagent
@Contact :   jerrychen1990@gmail.com
'''

from unittest import TestCase
# sys.path.append("../")
from xagents.agent.api import *
from xagents.kb.api import *
from loguru import logger
from snippets import set_logger

agent_name = "unittest_agent"

# unit test


class TestXagent(TestCase):

    @classmethod
    def setUpClass(cls):
        set_logger("dev", __name__)
        logger.info("start test embd")

    def test_list_agents(self):
        agents = list_agents()
        logger.info(f"agents: {agents}")
        agents_names = [agent.name for agent in agents]
        self.assertIn(agent_name, agents_names)

    def test_get_agent(self):
        agent: XAgent = get_agent(name=agent_name)
        logger.info(f"agent: {agent.get_info()}")

    def test_create_delete_agent(self):
        llm_config = dict(cls="GLM", version="glm-3-turbo")
        tools_config = [dict(name="计算器"), dict(name="车次查询")]
        agent_name = "temp_create_agent"
        agent = create_agent(name=agent_name, llm_config=llm_config,
                             kb_config=KBConfig(name="unittest_kb"),
                             tools_config=tools_config, store_type=STORE_TYPE.DISK)
        # logger.info(f"create Agent {agent.get_info()} success")
        agent = get_agent(name=agent_name)
        logger.info(f"create agent: {agent.get_info()} success")

        delete_agent(name=agent_name)
        logger.info(f"delete agent {agent_name} success")
        agent = get_agent(name=agent_name)

        self.assertRaises(Exception)

        # create agent

    def test_agent_tool_use(self):
        agent_name = "unittest_agent"
        llm_config = dict(cls="GLM", version="glm-3-turbo")
        tools_config = [dict(name="计算器"), dict(name="车次查询")]

        # create agent
        agent: XAgent = create_agent(name=agent_name, llm_config=llm_config, tools_config=tools_config)
        logger.info(f"create Agent {agent.get_info()} success")

        # chat with agent
        message = "你能帮我查询2024年1月1日从北京南站到上海的火车票吗？"

        llm_gen_config = LLMGenConfig()
        agent_resp = chat_agent(name=agent.name, message=message, stream=True, llm_gen_config=llm_gen_config)
        logger.info(f"agent resp: {agent_resp.model_dump(exclude={'content'})}")
        self.assertIsNotNone(agent_resp.content)
        for chunk in agent_resp.content:
            logger.info(chunk)
        self.assertIsNotNone(agent_resp.usage)

    def test_agent_tool_websearch_use(self):

        agent_name = "unittest_agent"
        llm_config = dict(cls="GLM", version="glm-3-turbo")
        kb_name = "unittest_kb"
        self._init_kb(kb_name)
        kb_config = KBConfig(name=kb_name)
        tools_config = [dict(name="联网查询")]
        #
        # web_search_config = WebSearchConfig(name = "test")
        # print(f"web_search_config1:{web_search_config}")

        logger.info(f"create tools_config_v1 {tools_config} success")
        # create agent
        agent: XAgent = create_agent(name=agent_name, llm_config=llm_config, tools_config=tools_config, kb_config=kb_config)
        logger.info(f"create Agent {agent.get_info()} success")

        # chat with agent
        message = "帮我查询一下杭州的最新天气"  # "你能帮我找一下今天关于杭州实时新闻吗？"
        kb_search_config = KBSearchConfig(do_expand=True, expand_len=200)
        llm_gen_config = LLMGenConfig()
        agent_resp = chat_agent(name=agent.name, message=message, use_kb=False, stream=False,
                                kb_search_config=kb_search_config, llm_gen_config=llm_gen_config)
        logger.info(f"agent resp: {agent_resp.model_dump()}")
        self.assertIsNotNone(agent_resp.content)
        self.assertIsNotNone(agent_resp.usage)

    def test_agent_multi_turn(self):
        agent_name = "unittest_agent"
        llm_config = dict(cls="GLM", version="glm-3-turbo")
        _system = "请用英语回答我的问题"

        # create agent
        agent: XAgent = create_agent(name=agent_name, llm_config=llm_config)
        logger.info(f"create Agent {agent.get_info()} success")

        # chat with agent
        message = "帮我推荐三首歌"
        llm_gen_config = LLMGenConfig(system=_system)
        agent_resp = chat_agent(name=agent.name, message=message, stream=True, llm_gen_config=llm_gen_config)
        # agent_resp = agent.chat(name=agent.name, message=message, llm_gen_config=llm_gen_config)

        self.assertIsNotNone(agent_resp.content)
        content = "".join(agent_resp.content)
        logger.info(content)
        logger.info(f"agent resp: {agent_resp.model_dump(exclude={'content'})}")

        self.assertIsNotNone(agent_resp.usage)

        # logger.debug(f"agent_memory:{agent.memory.to_llm_history()}")

        message = "详细介绍一下第三首"
        llm_gen_config = LLMGenConfig(system=_system)

        agent_resp = chat_agent(name=agent.name, message=message, stream=False, llm_gen_config=llm_gen_config, details=True)
        # agent_resp = agent.chat(name=agent.name, message=message, llm_gen_config=llm_gen_config,details=True)

        self.assertIsNotNone(agent_resp.content)
        logger.info(f"agent resp: {agent_resp.model_dump()}")

        self.assertIsNotNone(agent_resp.usage)
        self.assertNotEquals(len(agent_resp.details), 0)

    def _init_kb(self, kb_name: str):
        try:
            kb = get_knowledge_base(kb_name)
        except Exception as e:
            kb = create_knowledge_base(kb_name)
        file_name = "README.md"
        try:
            kb_file = get_kb_file_info(kb_name=kb_name, file_name=file_name)
        except Exception as e:
            cur_dir = os.path.abspath(os.path.dirname(__file__))
            file_path = os.path.join(os.path.dirname(cur_dir), file_name)
            logger.debug(f"adding file:{file_path}")
            kb_file = create_kb_file(kb_name=kb_name, file=file_path)

    def test_agent_kb(self):
        llm_config = dict(cls="GLM", version="glm-3-turbo")
        kb_name = "unittest_kb"
        self._init_kb(kb_name)
        kb_config = KBConfig(name=kb_name)

        # create agent
        agent: XAgent = create_agent(name=agent_name, llm_config=llm_config, kb_config=kb_config)
        logger.info(f"create Agent {agent.get_info()} success")

        # chat with agent
        message = "20240401发布了哪些功能？"
        llm_gen_config = LLMGenConfig()
        kb_search_config = KBSearchConfig(do_expand=True, expand_len=200)
        agent_resp = chat_agent(name=agent.name, message=message, use_kb=True, stream=False,
                                kb_search_config=kb_search_config, llm_gen_config=llm_gen_config)
        # agent_resp = agent.chat(name=agent.name, message=message, llm_gen_config=llm_gen_config)

        self.assertIsNotNone(agent_resp.content)
        content = "".join(agent_resp.content)
        logger.info(f"agent resp:{content}")

        self.assertIsNotNone(agent_resp.references)
        for reference in agent_resp.references:
            logger.info(f"reference_len:{reference.total_len}")
            logger.info(f"reference: {reference.to_plain_text()}")

        self.assertIsNotNone(agent_resp.usage)

        logger.info(f"agent resp: {agent_resp.model_dump(exclude={'content'})}")

    def test_inverted_index(self):
        # create kb
        kb_name = 'test_inverted'
        try:
            kb = get_knowledge_base(kb_name)
        except Exception as e:
            kb = create_knowledge_base(kb_name)
        file_name = "README.md"

        # create kb_file
        try:
            kb_file = get_kb_file_info(kb_name=kb_name, file_name=file_name)
        except Exception as e:
            cur_dir = os.path.abspath(os.path.dirname(__file__))
            file_path = os.path.join(os.path.dirname(cur_dir), file_name)
            logger.debug(f"adding file:{file_path}")
            kb_file = create_kb_file(kb_name=kb_name, file=file_path)

        # delete kb_file
        # cur_dir = os.path.abspath(os.path.dirname(__file__))
        # file_path = os.path.join(os.path.dirname(cur_dir), file_name)
        # delete_kb_file(kb_name=kb_name, file_name=file_name)

        # reindex_kb_file
        # reindex_kb_file(kb_name=kb_name, file_name=file_name, reindex=True)

        agent_name = "unittest_agent"
        kb_config = KBConfig(name=kb_name)
        llm_config = dict(cls="GLM", version="glm-3-turbo")

        # create agent
        agent: XAgent = create_agent(name=agent_name, llm_config=llm_config, kb_config=kb_config)
        logger.info(f"create Agent {agent.get_info()} success")

        # chat with agent
        message = "具体接口文档参考哪里？"
        llm_gen_config = LLMGenConfig()
        kb_search_config = KBSearchConfig(do_expand=True, expand_len=200)
        agent_resp = chat_agent(name=agent.name, message=message, use_kb=True, stream=False,
                                kb_search_config=kb_search_config, llm_gen_config=llm_gen_config)
        self.assertIsNotNone(agent_resp.content)
        content = "".join(agent_resp.content)
        logger.info(f"agent resp:{content}")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/08 14:38:44
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''

import copy
import re
from abc import abstractmethod
from typing import Iterable, List

from xagents.loader.common import ContentType
from xagents.config import *
from xagents.kb.common import Chunk
from snippets import batchify


class AbstractSplitter:
    def __init__(self, invalid_chunks: List[str] = []):
        self.invalid_chunks = invalid_chunks

    @abstractmethod
    def split(self, text: str) -> List[str]:
        raise NotImplementedError

    def split_chunk(self, chunk: Chunk) -> Iterable[Chunk]:
        if chunk.content_type == ContentType.IMAGE:
            # 图片、表格和 json 暂时不做切割
            yield chunk
        else:

            for content in self.split(chunk.content):
                # content = content.strip()
                if content in self.invalid_chunks:
                    continue
                yield Chunk(content=content,  content_type=chunk.content_type, page_idx=chunk.page_idx)

# 将文本切分、合并到（min_len~max_len）之间的长度


def merge_cut_texts(texts: Iterable[str], min_len: int, max_len: int) -> Iterable[str]:
    acc = ""
    for text in texts:
        acc += text
        if len(acc) < min_len:
            continue
        if len(acc) > max_len:
            for item in batchify(acc, max_len):
                item = "".join(item)
                if len(item) >= min_len:
                    yield item
                    acc = ""
                else:
                    acc = item
        else:
            yield acc
            acc = ""
    if acc:
        yield acc


class BaseSplitter(AbstractSplitter):

    def __init__(self,
                 separator="\n|。|？|\?|！|！|，",
                 parse_table=False,
                 max_len=100,
                 min_len=5,
                 **kwargs):
        super().__init__(**kwargs)
        self.parse_table = parse_table
        self.separator = separator
        self.max_len = max_len
        self.min_len = min_len

    def _parse_text(self, text: str) -> str:
        # text = text.strip()
        # text = re.sub("\s+", " ", text)
        # text = re.sub(f"\.{3,}", "", text)
        # text = text.strip()
        return text

    def split(self, text: str) -> Iterable[str]:
        # logger.info(f"splitting {text=}")

        # logger.info(f"{self.separator=}")

        texts = re.split(self.separator, text)
        texts = [self._parse_text(t) for t in texts]
        # logger.debug(f"cut text: {texts}")
        mcs = list(merge_cut_texts(texts, self.min_len, self.max_len))
        # logger.debug(f"merge cut text: {mcs}")
        yield from mcs


_SPLITTERS = [BaseSplitter]
_NAME2SPLITTER = {s.__name__: s for s in _SPLITTERS}


def get_splitter(config: dict) -> AbstractSplitter:
    tmp_config = copy.copy(config)
    splitter_cls = tmp_config.pop("splitter_cls")
    splitter_cls = _NAME2SPLITTER[splitter_cls]
    return splitter_cls(**tmp_config)


if __name__ == "__main__":
    # splitter = BaseSplitter()
    # texts = ["第一节  释义  ................................ ", " "]
    # for text in texts:
    #     print(splitter.split(text))

    # texts = ["a"*4, "b"*5, "b"*4, "c"*51, "d"*4]
    # for ele in merge_cut_texts(texts, min_len=5, max_len=25):
    #     print(f"{ele}, {len(ele)}")
    text = '''一二三四五六七八九十一二三四五六七八九十一二三四五六七八九十一二三四五六七八九十
一二三四五六七八九十

一二三四五六七八九十

一二三四五六七八九十

一二三四五六七八九十


一二三四五六七八九十

一二三四五六七八九十


一二三四五六七八九十'''

    for ele in merge_cut_texts([text], min_len=1, max_len=10):
        print(f"{ele}, {len(ele)}")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/03/01 11:10:10
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''


from loguru import logger
from xagents.config import *
from snippets import print_info, set_logger



    
set_logger(env = XAGENT_ENV, module_name=__name__, log_dir=LOG_DIR)
set_logger(XAGENT_ENV, "__main__")


def show_env():
    print_info("current XAgent env", logger)
    logger.info(f"{XAGENT_ENV=}")
    logger.info(f"{KNOWLEDGE_BASE_DIR=}")
    logger.info(f"{TEMP_DIR=}")
    logger.info(f"{LOG_DIR=}")
    logger.info(f"{AGENT_DIR=}")
    logger.info(f"{DATA_DIR=}")
    print_info("", logger)

show_env()





#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/07 17:57:13
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''


import os
import copy
from typing import Generator, List, Optional

from pydantic import BaseModel, Field

from agit.common import LLMResp
from xagents.util import save_format
from xagents.model.common import LLMGenConfig
from xagents.agent.common import AbstractAgent, AgentResp, KBConfig, WebSearchResp
from xagents.config import LOG_DIR
from xagents.kb.common import RecalledChunk
from xagents.kb.api import KBSearchConfig, get_knowledge_base
from xagents.model.api import get_llm_model
from xagents.agent.memory import BaseMemory
from xagents.tool.api import get_tools, invoke_tool_call
from xagents.tool.core import BaseTool, ToolCall, ToolDesc
from snippets import dump, jload

from loguru import logger

agent_log_path = os.path.join(LOG_DIR, "xagent.log")


class XAgentInfo(BaseModel):
    name: str = Field(description="agent的名称，唯一键")
    llm_config: dict = Field(description="llm的配置信息")
    memory_config: dict = Field(description="memory的配置信息")
    kb_config: Optional[KBConfig] = Field(description="kb的配置信息")
    # web_search_config: Optional[WebSearchConfig] = Field(description="web_search的配置信息")
    tools: List[ToolDesc] = Field(description="工具列表")


class XAgent(AbstractAgent):

    def __init__(self, name: str,
                 llm_config: dict,
                 memory_config: dict,
                 kb_config: KBConfig,
                 # web_search_config: WebSearchConfig,
                 tools: List[BaseTool] = []) -> None:
        super().__init__(name=name)
        self.info = XAgentInfo(name=name, llm_config=llm_config, memory_config=memory_config, kb_config=kb_config,
                               tools=[ToolDesc(**e.model_dump(exclude={})) for e in tools])
        self.llm_model = get_llm_model(llm_config)
        self.memory = BaseMemory(**memory_config)
        self._load_kb(kb_config)
        # print(f"web_search_config_0:{web_search_config}")
        # self._load_web_search(web_search_config)
        self.tools = tools

    def _load_kb(self, kb_config: KBConfig):
        if not kb_config:
            self.kb = None
            self.kb_prompt_template = None
        else:
            self.kb = get_knowledge_base(kb_config.name)
            self.kb_prompt_template = kb_config.prompt_template
            logger.info("load kb finish")

    def get_info(self):
        return self.info

    def search_kb(self, query: str, **kwargs) -> List[RecalledChunk]:
        chunks = self.kb.search(query=query, **kwargs)
        return chunks

    def use_tool(self, tool_call: ToolCall):
        resp = invoke_tool_call(tool_call)
        tool_call.resp = resp
        return resp

    @staticmethod
    def _get_agent_config_path(dir_path: str, agent_name: str):
        return os.path.join(dir_path, agent_name+".json")

    def save(self, save_dir: str):
        save_path = self._get_agent_config_path(save_dir, self.name)
        logger.info(f"save agent:{self.name} to {save_path}")
        logger.debug(f"agent info:{self.get_info()}")
        dump(self.get_info().model_dump(), save_path)

    def delete(self, save_dir: str):
        save_path = self._get_agent_config_path(save_dir, self.name)
        if os.path.exists(save_path):
            logger.info(f"deleting agent:{self.name} from save_path:{save_path}")
            os.remove(save_path)

    @classmethod
    def from_config(cls, config: str | dict) -> "XAgent":
        config = jload(config) if isinstance(config, str) else config
        if config.get("tools"):
            tools = get_tools(config["tools"])
            config["tools"] = tools
        if config.get("kb_config"):
            config["kb_config"] = KBConfig.model_validate(config["kb_config"])

        return XAgent(**config)

    def chat(self, message: str, history: List[dict] = None, do_remember=True, details=False, stream=False,
             use_kb=False, kb_search_config: KBSearchConfig = KBSearchConfig(),
             fake_chat=False, llm_gen_config: LLMGenConfig = LLMGenConfig(), **kwargs) -> AgentResp:

        chunks = []
        if use_kb:
            if not self.kb or not self.kb_prompt_template:
                logger.warning(f"agent:{self.name} has no related knowledge base, will not chat with kb! ")
                prompt = message
                chunks = None
            else:
                logger.info("agent searching kb with kb_name")
                chunks = self.search_kb(query=message, **kb_search_config.model_dump())
                context = "\n".join(f"{idx+1}." + c.to_plain_text() for idx, c in enumerate(chunks))
                prompt = save_format(template=self.kb_prompt_template, question=message, context=context)

                # prompt = self.kb_prompt_template.format(question=message, context=context)
        else:
            prompt = message
            chunks = None

        web_search_result = None
        tool_calls = None
        if fake_chat:
            fake_resp = "这是MOCK的回答信息,如果需要真实回答,请设置fake_chat=False"
            llm_resp = (e for e in fake_resp) if stream else fake_resp
            tool_call = None
        else:
            history = history if history is not None else self.memory.to_llm_history()
            llm_resp: LLMResp = self.llm_model.generate(prompt=prompt, history=history, tools=self.tools, details=details, stream=stream,
                                                        **llm_gen_config.model_dump(), **kwargs)

            logger.debug(f"llm_resp_inner:{llm_resp}")
            # 调用tool
            tool_calls = copy.deepcopy(llm_resp.tool_calls)
            if tool_calls:
                for tool_call in tool_calls:
                    logger.debug(f"calling tool:{tool_call.name}")
                    tool_resp = self.use_tool(tool_call)
                    logger.debug(f"tool_resp:{tool_resp}")
                    llm_resp = self.llm_model.observe(prompt=prompt, tool_call=tool_call, tools=self.tools, history=history, details=details, stream=stream,
                                                      **llm_gen_config.model_dump(), **kwargs)
                    if tool_call.name == "联网查询":
                        search_results, sub_page_contents = tool_resp[0], tool_resp[1]
                        search_info = "\n".join(sub_page_contents)
                        web_search_result = WebSearchResp(summary=search_info, pages=None)

        def _remember_callback(resp_str):
            if do_remember:
                self.remember("user", message)
                self.remember("assistant", resp_str)

        def _add_remember_callback(gen: Generator) -> Generator:
            acc = []
            for ele in gen:
                acc.append(ele)
                yield ele
            resp_str = "".join(acc)
            logger.info(f"agent generate response:{resp_str}")
            _remember_callback("".join(acc))

        if stream:
            content = _add_remember_callback(llm_resp.content)
        else:
            content = llm_resp.content
            logger.info(f"generate response:{content}")
            _remember_callback(content)

        resp = AgentResp(content=content, tool_calls=tool_calls, usage=llm_resp.usage, references=chunks,
                         details=llm_resp.details, web_search_result=web_search_result)
        return resp

    def remember(self, role: str, message: str):
        logger.debug(f"remembering {role=}, {message=}")
        self.memory.remember(role, message)

    def clear_memory(self):
        # logger.info("clearing memory")
        self.memory.clear()

    def __repr__(self) -> str:
        return f"XAgent-{self.name}"


if __name__ == "__main__":
    llm_config = dict(model_cls="GLM", name="glm", version="chatglm_turbo")
    memory_config = dict(size=10)
    agent = XAgent(name="xagent", llm_config=llm_config, memory_config=memory_config)
    resp = agent.chat("推荐三首歌", stream=True)
    for item in resp:
        print(item, end="")
    resp = agent.chat("其中第二首是谁唱的？", stream=True)
    for item in resp:
        print(item, end="")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import os
import pptx
import io
import csv
import pandas as pd

from PIL import Image
from typing import Iterable
from xagents.loader.utils import upload_to_minio, image2text
from xagents.loader.common import Chunk, AbstractLoader, ImageChunk, TextChunk, TableChunk, get_image_name_path, merge_chunks


def read_pptx_tab(tab, **kwargs):
    vf = io.StringIO()
    writer = csv.writer(vf)
    for row in tab.rows:
        writer.writerow(cell.text for cell in row.cells)
    vf.seek(0)
    return pd.read_csv(vf, **kwargs)


def analyse_slide(powerpoint:pptx.Presentation, file_name:str, upload_image:bool, ocr:bool)->Iterable[Chunk]:
    page_index = 1
    image_index = 0

    for slide in powerpoint.slides:
        for shape in slide.shapes:
            if shape.has_text_frame:  # text
                content = shape.text_frame.text
                yield TextChunk(content=content, page_idx=page_index)
            
            elif shape.has_table:  # table
                df:pd.DataFrame = read_pptx_tab(shape.table)
                yield TableChunk(page_idx=page_index, data=df)
            
            elif shape.shape_type == 13:  # 图片类型，MSO_SHAPE_TYPE.PICTURE
                image_name, image_path = get_image_name_path(file_name, page_index, image_index)
                if upload_image:
                    with open(image_path, "wb") as f:
                        image = shape.image
                        image_bytes = image.blob
                        f.write(image_bytes)
                    url = upload_to_minio(image_path)
                else:
                    url=None
                content = image2text(Image.open(io.BytesIO(shape.image.blob))) if ocr else None
                image_index += 1
                yield ImageChunk(url=url, content=content,page_idx=page_index, image_name=image_name)
        page_index += 1


class PPTXLoader(AbstractLoader):
    def __init__(self, max_page:int=None, **kwargs):
        super().__init__(**kwargs)
        self.max_page = max_page
    
    def load(self, file_path: str, start_page=1, end_page=None, upload_image=True, ocr=False, **kwargs) -> Iterable[Chunk]:
        ppt = pptx.Presentation(file_path)
        chunks = analyse_slide(ppt,  file_name=os.path.basename(file_path), upload_image=upload_image, ocr=ocr)
        chunks = merge_chunks(chunks)
        yield from chunks


if __name__ == "__main__":
    ppt = pptx.Presentation("data/kb_file/时序数据预测.pptx")
    chunks = analyse_slide(ppt)
    for chunk in chunks:
        print(chunk)


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/11 16:30:03
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''


import os
import re
from typing import Iterable
from xagents.loader.common import AbstractLoader, Chunk, TextChunk
from loguru import logger


def extract_md_table(text):
    tables = re.findall("^\\|(.*)\\|$", text, re.MULTILINE | re.DOTALL)
    return tables


class MarkDownLoader(AbstractLoader):
    def load(self, file_path: str, **kwargs) -> Iterable[Chunk]:
        with open(file_path, "r") as f:
            content = f.read()
            # logger.debug(f"{content=}")
            yield TextChunk(content=content, page_idx=1)


if __name__ == "__main__":
    from xagents.config import XAGENT_HOME
    loader = MarkDownLoader()

    file_path = os.path.join(XAGENT_HOME, "data/raw/贵州茅台2022年报-4.md")
    chunks = loader.load(file_path=file_path)

    print(len(chunks))
    print(chunks[0])



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/04/25 13:55:58
@Author  :   ChenHao
@Description  :   load过程中的工具类
@Contact :   jerrychen1990@gmail.com
'''

from loguru import logger
from minio import Minio
from xagents.config import *



def create_minio_client():
    return Minio(MINIO_URL,
                 access_key=MINIO_ACCESS_KEY,
                 secret_key=MINIO_SECRET_KEY,
                 secure=False)


def upload_to_minio(file_path, bucket_name="xagent") -> str:
    try:
        # 使用 fput_object 上传文件
        minioClient = create_minio_client()
        object_name = os.path.basename(file_path)
        minioClient.fput_object(bucket_name, object_name, file_path)        
        logger.info(f'Successfully uploaded {object_name} to {bucket_name}/{object_name}')
        url = minioClient.presigned_get_object(bucket_name, object_name)
        return url
    except Exception as err:
        logger.exception(err)
        return None
    

def image2text(image_path):
    from cnocr import CnOcr
    ocr = CnOcr(det_model_name='ch_PP-OCRv3_det')  # model_name 参考 https://gitee.com/cyahua/cnocr
    text = ocr.ocr(image_path)
    content = ','.join(item['text'] for item in text)
    return content


if __name__ == "__main__":
    file_path = os.path.join(DATA_DIR, "image_table_pages-page1-image0.png")
    url = upload_to_minio(file_path)
    print(url)
   



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/03/19 16:24:43
@Author  :   ChenHao
@Description  : 文档处理对外接口
@Contact :   jerrychen1990@gmail.com
'''

import os
from loguru import logger
from typing import Iterable, List, Type
from xagents.loader.pdf_loader import PDFLoader
from xagents.loader.docx_loader import DOCXLoader
from xagents.loader.doc_loader import DOCLoader
from xagents.loader.csv_loader import CSVLoader
from xagents.loader.excel_loader import EXCELLoader
from xagents.loader.xls_loader import XLSLoader
from xagents.loader.pptx_loader import PPTXLoader
from xagents.loader.ppt_loader import PPTLoader

from xagents.loader.markdown_loader import MarkDownLoader
from xagents.loader.json_loader import JSONLoader
from xagents.loader.jsonl_loader import JSONLLoader
from xagents.loader.common import Chunk, AbstractLoader

from xagents.loader.splitter import BaseSplitter
from snippets import flat, log_cost_time

_EXT2LOADER = {
    "pdf": PDFLoader,
    "markdown": MarkDownLoader,
    "md": MarkDownLoader,
    "json": JSONLoader,
    "jsonl": JSONLLoader,
    "csv": CSVLoader,
    "xlsx": EXCELLoader,
    "xls": XLSLoader,
    "txt": MarkDownLoader,
    "docx": DOCXLoader,
    "doc": DOCLoader,
    "pptx": PPTXLoader,
    "ppt": PPTLoader,
    "": MarkDownLoader
}


def get_loader_cls(file_path: str) -> Type[AbstractLoader]:
    """根据file路径后缀，加载对应的文档加载器

    Args:
        file_path (str): 文件后缀

    Raises:
        ValueError: 非法后缀异常

    Returns:
        Type[AbstractLoader]: 文档加载器class
    """
    ext = os.path.splitext(file_path)[-1].lower().replace(".", "")
    if ext not in _EXT2LOADER:
        msg = f"{ext} not support!"
        raise ValueError(msg)

    loader_cls = _EXT2LOADER[ext]
    return loader_cls


@log_cost_time(name="load file", logger=logger)
def load_file(file_path: str, start_page: int = 1, end_page: int = None, upload_image=True, ocr=False, **kwargs) -> Iterable[Chunk]:
    """处理文档，返回文档切片

    Args:
        file_path (str): 文档路径
        start_page (int, optional): 开始页码. Defaults to 1.
        end_page (int, optional): 结束页码,None表示到最后一页结束. Defaults to None.
        upload_image (bool, optional): 是否存储图片并上传到图片服务器. Defaults to True.
        ocr (bool, optional): 是否使用ocr识别图片中的文字. Defaults to False.

    Returns:
        Iterable[Chunk]: 切片列表
    """
    loader_cls = get_loader_cls(file_path)
    loader: AbstractLoader = loader_cls(**kwargs)
    logger.debug(f"loading {file_path} with loader:{loader}")
    pages = loader.load(file_path, start_page=start_page, end_page=end_page, upload_image=upload_image, ocr=ocr,
                        **kwargs)
    return pages


def convert2txt(file_path: str, dst_path: str = None, **kwargs) -> str:
    """将原始文件转移成txt格式

    Args:
        file_path (str): 原始文件路径
        dst_path (str, optional): 目标路径，未传的话，和原始文件同目录. Defaults to None.

    Returns:
        str: 目标路径
    """
    chunks = load_file(file_path, **kwargs)
    if not dst_path:
        dst_path = file_path+".txt"
    with open(dst_path, 'w', encoding='utf-8') as f:
        for chunk in chunks:
            f.write(chunk.content)
    return dst_path


def parse_file(file_path: str,
               start_page: int = 1,
               end_page: int = None,
               upload_image=True,
               ocr=False,
               do_cut: bool = True,
               separator: str = '\n',
               max_len: int = 200,
               min_len: int = 10) -> List[Chunk]:
    """加载并切分切片

    Args:
        file_path (str): 文件路径
        start_page (int, optional): 开始页码. Defaults to 1.
        end_page (int, optional): 结束页码. Defaults to None.
        upload_image (bool, optional): 是否上传图片到存储服务器. Defaults to True.
        ocr (bool, optional): 是否调用ocr. Defaults to False.
        do_cut (bool, optional): 是否切片. Defaults to True.
        separator (str, optional): 分隔符. Defaults to '\n'.
        max_len (int, optional): 最大切片长度. Defaults to 200.
        min_len (int, optional): 最小切片长度. Defaults to 10.

    Returns:
        List[Chunk]: 切片列表
    """

    splitter = BaseSplitter(max_len=max_len, min_len=min_len, separator=separator)
    # logger.debug(f"splitter: {splitter}")
    origin_chunks: Iterable[Chunk] = load_file(file_path=file_path, start_page=start_page,
                                               end_page=end_page, upload_image=upload_image, ocr=ocr)
    origin_chunks = list(origin_chunks)
    logger.debug(f"load {len(origin_chunks)} origin_chunks")
    # logger.debug(f"{origin_chunks[0]=}")
    if do_cut:
        split_chunks = flat([splitter.split_chunk(origin_chunk) for origin_chunk in origin_chunks])
        logger.info(f"split {len(origin_chunks)} origin_chunks to {len(split_chunks)} chunks")
        return split_chunks
    else:
        return origin_chunks


if __name__ == "__main__":
    file_path = "/Users/chenhao/Downloads/汽车手册V16.txt"

    chunks = parse_file(file_path, do_cut=True, separator="==")
    print(f"{len(chunks)=}")
    for chunk in chunks[:4]:
        print(chunk)
        # print(chunk.content)



import os
from typing import Iterable
from xagents.loader.common import AbstractLoader, Chunk, merge_chunks
from xagents.loader.excel_loader import analyse_table
import subprocess

def xls2xlsx(xls_path, xlsx_path):
    # 构建命令行命令
    cmd = ["soffice", "--headless", "--convert-to", "xlsx", "--outdir", xlsx_path, xls_path]
    # 执行命令
    subprocess.run(cmd)
    print(f"Converted '{xls_path}' to xlsx format")

class XLSLoader(AbstractLoader):
    def __init__(self, max_page:int=None, **kwargs):
        super().__init__(**kwargs)
        self.max_page = max_page
    
    def load(self, file_path: str, **kwargs) -> Iterable[Chunk]:
        directory = os.path.dirname(file_path)
        xls2xlsx(file_path, directory)
        chunks = analyse_table(f"{file_path}x", **kwargs)
        chunks = merge_chunks(chunks)
        # os.remove(f"{file_path}x")  # 删除产生的中间文件 xlsx
        yield from chunks


if __name__ == "__main__":
    from xagents.config import XAGENT_HOME
    loader = XLSLoader()

    file_path = os.path.join(XAGENT_HOME, "大学综合排名2022.xls")
    chunks = loader.load(file_path=file_path)

    print(len(chunks))
    print(chunks[0])


#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import os
import jsonlines
from typing import Iterable
from xagents.loader.common import AbstractLoader, Chunk, merge_chunks
from xagents.loader.json_loader import analyse_json

class JSONLLoader(AbstractLoader):
    def load(self, file_path: str, upload_image=True, ocr=False, **kwargs) -> Iterable[Chunk]:
        with open(file_path, "r", encoding='utf-8') as f:
            content = jsonlines.Reader(f)
        
            chunks = analyse_json(file_path=file_path, content=content, upload_image=upload_image, ocr=ocr)
            yield from chunks



if __name__ == "__main__":
    from xagents.config import XAGENT_HOME
    loader = JSONLLoader()

    file_path = os.path.join(XAGENT_HOME, "")
    chunks = loader.load(file_path=file_path)

    print(len(chunks))
    print(chunks[0])



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import os
import json
from typing import Iterable
from xagents.loader.common import AbstractLoader, Chunk, ImageChunk, TextChunk, TableChunk, merge_chunks

def analyse_json(file_path:str, content, upload_image:bool, ocr:bool)->Iterable[Chunk]:
    for item in content:
        if not isinstance(item, dict):
            raise Exception(f"Parsing file {file_path} failed: unexpected json format.")
        try:
            if item['content_type'] == "TEXT":
                yield TextChunk(content=item['content'], page_idx=item['page_idx'])
            elif item['content_type'] == "IMAGE":
                yield ImageChunk(**item)
            elif item['content_type'] == "TABLE":
                yield TableChunk(**item)
        except:
            raise IndexError()

class JSONLoader(AbstractLoader):
    def load(self, file_path: str, upload_image=True, ocr=False, **kwargs) ->Iterable[Chunk]:
        with open(file_path, "r", encoding='utf-8') as f:
            content = json.load(f)
            if not isinstance(content, list):
                raise Exception(f"Parsing file {file_path} failed: unexpected json format.")
        
        chunks = analyse_json(file_path=file_path, content=content, upload_image=upload_image, ocr=ocr)
        yield from chunks
        


if __name__ == "__main__":
    from xagents.config import XAGENT_HOME
    loader = JSONLoader()

    file_path = os.path.join(XAGENT_HOME, "alpaca_data-0-3252-中文.json")
    chunks = loader.load(file_path=file_path)

    print(len(chunks))
    print(chunks[0])



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/01/04 14:15:58
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''

from xagents.tool.common_tool import *
from xagents.tool.core import *










#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/07 17:45:15
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''


from abc import abstractmethod
from typing import Generator, Optional, Union

from agit.llm import LLMResp
from agit.common import ToolCall

from langchain_core.embeddings import Embeddings
from pydantic import BaseModel, Field


StrOrGen = Union[str, Generator]


class LLMGenConfig(BaseModel):
    system: Optional[str] = Field(description="系统信息", default=None)
    temperature: float = Field(default=0.95, ge=0, le=1, description="生成文本的温度")
    top_p: float = Field(default=0.7, ge=0, le=1, description="生成文本的top_p")
    max_tokens: int = Field(default=2048, ge=1, description="生成文本的最大长度")
    do_sample: bool = Field(default=True, description="是否使用采样")


class LLMGen(BaseModel):
    content: StrOrGen = Field(description="生成的文本内容, 字符串或生成器")
    tool_call: Optional[ToolCall] = Field(description="工具调用")
    detail: dict = Field()


class LLM:
    """语言模型的基类
    """

    def __init__(self, version: str, name: str = None):
        """初始化函数

        Args:
            name (str): LLM的名称
            version (str): 模型版本
        """
        self.name = name if name else self.__class__.__name__ + "-" + version
        self.version = version

    @abstractmethod
    def generate(self, prompt: str, history=[], system: str = None, stream=True, **kwargs) -> LLMResp:
        """生成结果

        Args:
            prompt (str): 给LLM的提示词
            history (list, optional): 历史message列表. Defaults to [].
            system (_type_, optional): system信息. Defaults to None.
            stream (bool, optional): 是否返回generator. Defaults to True.
        Returns:
            Tuple[ToolCall,StrOrGen]: ToolCall:工具调用实例，可以为空, StrOrGen:模型回复内容，字符串或generator
        """
        raise NotImplementedError

    @classmethod
    def list_version(cls):
        raise NotImplementedError

    def __str__(self):
        return f"[{self.name}]-{self.version}"

    def __repr__(self):
        return self.__str__()


class EMBD(Embeddings):
    def get_dim(self) -> int:
        raise NotImplementedError


class Reranker:
    def __init__(self, name: str, version: str):
        self.name = name
        self.version = version

    @abstractmethod
    def cal_similarity(self, text1: str, text2: str) -> float:
        raise NotImplementedError



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import os
import docx
from typing import Iterable
from xagents.loader.common import AbstractLoader, Chunk, merge_chunks
from xagents.loader.docx_loader import analyse_page
import subprocess

def doc2docx(doc_path, docx_path):
    # 构建命令行命令
    cmd = ["soffice", "--headless", "--convert-to", "docx", "--outdir", docx_path, doc_path]
    # 执行命令
    subprocess.run(cmd)
    print(f"Converted '{doc_path}' to docx format")


class DOCLoader(AbstractLoader):
    def __init__(self, max_page:int=None, **kwargs):
        super().__init__(**kwargs)
        self.max_page = max_page
    
    def load(self, file_path: str, upload_image=True, ocr=False, **kwargs) ->Iterable[Chunk]:
        directory = os.path.dirname(file_path)
        doc2docx(file_path, directory)
        doc = docx.Document(f"{file_path}x")
        chunks = analyse_page(doc, file_name=os.path.basename(file_path), upload_image=upload_image, ocr=ocr)
        chunks = merge_chunks(chunks)
        os.remove(f"{file_path}x")  # 删除产生的中间文件 docx
        yield from chunks
        


if __name__ == "__main__":
    from xagents.config import XAGENT_HOME
    loader = DOCLoader()

    file_path = os.path.join(XAGENT_HOME, "初步设计文件.doc")
    chunks = loader.load(file_path=file_path)

    print(len(chunks))
    print(chunks[0])



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/04/12 15:02:40
@Author  :   ChenHao
@Description  :   测试loader
@Contact :   jerrychen1990@gmail.com
'''
from unittest import TestCase
from loguru import logger
from snippets import set_logger
from xagents.loader.api import load_file, parse_file
from xagents.config import *
from xagents.loader.common import *


# unit test
class TestLoader(TestCase):

    @classmethod
    def setUpClass(cls):
        set_logger("dev", __name__)
        logger.info("start test embd")

    @classmethod
    def show_chunks(cls, chunks: Iterable[Chunk]):
        image_cnt = 0
        for chunk in chunks:
            logger.info(f"page={chunk.page_idx}, type={chunk.content_type}")

            if chunk.content_type == ContentType.IMAGE:
                image_cnt += 1
                logger.info(f"image:{chunk.image_name}")
                logger.info(f"content:{chunk.content}")
                logger.info(f"url:{chunk.url}")
            if chunk.content_type == ContentType.TEXT:
                logger.info(f"text:{chunk.content}")

            if chunk.content_type == ContentType.TABLE:
                # logger.info(f"table:\n{chunk.data}")
                logger.info(f"content:\n{chunk.content}")
            logger.info("*"*40+"\n")
        return image_cnt

    def test_load_pdf(self):
        file_path = os.path.join(DATA_DIR, "kb_file", "image_table.pdf")
        chunks = load_file(file_path, end_page=5, ocr=False)
        # image_cnt = 0
        image_cnt = self.show_chunks(chunks)

        self.assertEqual(image_cnt, 4)

    def test_load_docx(self):
        file_path = os.path.join(DATA_DIR, "kb_file", "Xagents中间件.docx")
        chunks = load_file(file_path, ocr=False)
        image_cnt = self.show_chunks(chunks)

        self.assertEqual(image_cnt, 1)

    # def test_load_doc(self):
    #     file_path = os.path.join(DATA_DIR, "kb_file", "初步设计文件.doc")
    #     chunks = load_file(file_path)
    #     image_cnt = self.show_chunks(chunks)

    #     self.assertEqual(image_cnt, 6)

    def test_load_pptx(self):
        file_path = os.path.join(DATA_DIR, "kb_file", "时序数据预测.pptx")
        chunks = load_file(file_path)
        image_cnt = self.show_chunks(chunks)

        self.assertEqual(image_cnt, 18)

    def test_load_json(self):
        file_path = os.path.join(DATA_DIR, "kb_file", "alpaca_data-0-3252-中文.json")
        chunks = list(load_file(file_path))
        self.show_chunks(chunks)
        self.assertEqual(len(chunks), 7)

    def test_load_jsonl(self):
        file_path = os.path.join(DATA_DIR, "kb_file", "LICENSE.jsonl")
        chunks = list(load_file(file_path))
        self.show_chunks(chunks)
        self.assertEqual(len(chunks), 17)

    def test_load_excel(self):
        file_path = os.path.join(DATA_DIR, "kb_file", "大学综合排名2022.xlsx")
        chunks = list(load_file(file_path))

        self.show_chunks(chunks)

        self.assertEqual(len(chunks), 3)

    def test_load_txt(self):
        file_path = os.path.join(DATA_DIR, "kb_file", "requirements.txt")
        chunks = load_file(file_path)
        image_cnt = self.show_chunks(chunks)
        self.assertEqual(image_cnt, 0)
        chunks = parse_file(file_path=file_path)
        logger.info(chunks)
        self.show_chunks(chunks)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/03/19 16:10:35
@Author  :   ChenHao
@Description  :  知识库文件类
@Contact :   jerrychen1990@gmail.com
'''


# 知识库文件类
import os
from loguru import logger
from typing import List
from xagents.kb.common import KnowledgeBaseFileInfo, get_chunk_path, get_origin_path, get_status_path
from xagents.loader.api import parse_file
from xagents.loader.common import Chunk
from snippets import dump, load, jload, jdump


class KnowledgeBaseFile:
    def __init__(self, kb_name: str, file_name: str):
        self.kb_name = kb_name
        self.file_name = file_name
        self.chunk_path = get_chunk_path(self.kb_name, self.file_name)
        self.origin_path = get_origin_path(self.kb_name, self.file_name)

    @property
    def is_cut(self) -> bool:
        return os.path.exists(self.chunk_path)

    @property
    def chunk_num(self) -> int:
        if not self.is_cut:
            return 0
        chunks = load(self.chunk_path)
        return len(chunks)

    def set_index_status(self, is_indexed: bool):
        logger.debug(f"setting {self.file_name}'s is_indexed value to {is_indexed}")
        status_path = get_status_path(self.kb_name)
        status = jload(status_path)
        if self.file_name not in status:
            status[self.file_name] = {}
        status[self.file_name]["indexed"] = is_indexed
        # logger.debug(f"{status=}")
        jdump(status, status_path)

    @property
    def is_indexed(self) -> bool:
        status_path = get_status_path(self.kb_name)
        rs = jload(status_path).get(self.file_name, {}).get("indexed", False)
        return rs

    def get_info(self):
        return KnowledgeBaseFileInfo(kb_name=self.kb_name, file_name=self.file_name, is_cut=self.is_cut, chunk_num=self.chunk_num, is_indexed=self.is_indexed)

    def cut(self, *args, **kwargs) -> List[Chunk]:
        logger.info(f"start cut file: {self.file_name}")
        chunks: List[Chunk] = parse_file(file_path=self.origin_path, *args, **kwargs)
        self._save_chunks(chunks=chunks)
        return len(chunks)

    def _save_chunks(self, chunks: List[Chunk]):
        chunk_json = [chunk.to_json() for chunk in chunks]
        dump(chunk_json, self.chunk_path)
        return chunks

    def list_chunks(self) -> List[Chunk]:
        """
        从切片文件加载切片
        """
        if not self.is_cut:
            return []

        chunk_dicts = load(self.chunk_path)
        logger.debug(f"loaded {len(chunk_dicts)} chunks from {self.chunk_path}")
        chunks = [Chunk.from_dict(ele) for ele in chunk_dicts]
        return chunks

    def get_chunk_ids(self) -> List[str]:
        chunks = self.list_chunks()
        chunk_ids = [c.id for c in chunks]
        return chunk_ids

    def get_chunk(self, idx: int) -> Chunk:
        chunks = self.list_chunks()
        if idx < 0 or idx >= len(chunks):
            raise Exception(f"chunk index out of range, idx: {idx}, len: {len(chunks)}")
        return chunks[idx]

    def add_chunks(self, chunks: List[Chunk], idx: int = None):
        origin_chunks = self.list_chunks()
        if idx is None:
            idx = len(chunks)
        chunks = origin_chunks[:idx] + chunks + origin_chunks[idx:]
        self._save_chunks(chunks)

    def delete_chunks(self, ids: List[str]):
        chunks = self.list_chunks()
        remain_chunks = [c for c in chunks if c.id not in ids]
        self._save_chunks(remain_chunks)

    def update_chunk(self, chunk: Chunk) -> bool:
        # do delete

        chunks = self.list_chunks()
        for idx, c in enumerate(chunks):
            if c.id == chunk.id:
                chunks[idx] = chunk
                break
        self._save_chunks(chunks)

    def delete(self):
        """
        删除知识库文档
        """
        for path in [self.origin_path, self.chunk_path]:
            if os.path.exists(path):
                os.remove(path)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/07 18:33:25
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''


from abc import abstractmethod
import copy


class AbstractMemory:
    @abstractmethod
    def remember(self, role: str, content: str):
        raise NotImplementedError

    @abstractmethod
    def to_llm_history(self):
        raise NotImplementedError

    def clear(self):
        raise NotImplementedError


class BaseMemory(AbstractMemory):
    def __init__(self, size: int):
        self.size = size
        self.memory = []

    def remember(self, role: str, content: str):
        self.memory.append(dict(role=role, content=content))

    def to_llm_history(self):
        return copy.copy(self.memory)

    def clear(self):
        self.memory.clear()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/04/01 13:46:25
@Author  :   ChenHao
@Description  :   Agent问答的api
@Contact :   jerrychen1990@gmail.com
'''
from enum import Enum
import os
from typing import List

from xagents.tool.api import get_tools
from xagents.model.common import LLMGenConfig
from xagents.kb.api import KBSearchConfig
from xagents.config import AGENT_DIR, XAGENT_CACHE_NUM
from xagents.agent.common import AgentResp, KBConfig
from xagents.agent.xagent import XAgent
from cachetools import LRUCache, cached
from loguru import logger

agent_cache = LRUCache(maxsize=XAGENT_CACHE_NUM)


class STORE_TYPE(str, Enum):
    CACHE = "CACHE"
    DISK = "DISK"


def create_agent(name: str, llm_config: dict = dict(cls="GLM", name="glm", version="glm-4"),
                 tools_config: List[dict] = [], memory_config: dict = dict(size=10),
                 kb_config: KBConfig = None, store_type=STORE_TYPE.CACHE) -> XAgent:

    logger.info(f"creating agent:{name}")
    # print(f"web_search_config2:{web_search_config}")

    @cached(agent_cache)
    def _get_or_create(name: str):
        tools = get_tools(tools_config)  # [{'name': '联网查询'}]
        # print(f"web_search_config3:{web_search_config}")
        logger.info(f"getting tools_config:{tools_config}")
        logger.info(f"getting tools:{tools}")
        xagent = XAgent(name=name, llm_config=llm_config, memory_config=memory_config, kb_config=kb_config, tools=tools)
        return xagent
    agent = _get_or_create(name=name)
    if store_type == STORE_TYPE.DISK:
        agent.save(AGENT_DIR)

    return agent


@cached(agent_cache)
def get_agent(name: str) -> XAgent:
    agent = XAgent.from_config(os.path.join(AGENT_DIR, name+".json"))
    return agent


def chat_agent(name: str, message: str, history: List[dict] = None, do_remember=True, details=False, stream=False,
               use_kb=False, kb_search_config: KBSearchConfig = KBSearchConfig(),
               fake_chat=False, llm_gen_config: LLMGenConfig = LLMGenConfig()) -> AgentResp:
    """和Agent对话

    Args:
        name (str): Agent的名称
        message (str): 用户发送的消息
        history (List[dict], optional): 历史对话记录. Defaults to None.
        do_remember (bool, optional): Agent是否要记住历史对话. Defaults to True.
        details (bool, optional): 是否返回详细调用信息. Defaults to False.
        use_kb (bool, optional): 是否使用知识库. Defaults to False.
        kb_search_config (KBSearchConfig, optional): 知识库查询参数，use_kb=True时生效. Defaults to KBSearchConfig().
        fake_chat (bool, optional): 是否fake大模型问答（fake_chat=True时，不会调用LLM）. Defaults to False.
        llm_gen_config (LLMGenConfig, optional): 大模型生成文本需要的参数. Defaults to LLMGenConfig.

    Returns:
        AgentResp: Agent回答的结构体
    """
    # logger.info(f"sending message:{message} to agent:{name}")
    agent = get_agent(name=name)
    agent_resp = agent.chat(message=message, history=history, do_remember=do_remember, details=details, stream=stream,
                            use_kb=use_kb, kb_search_config=kb_search_config,
                            fake_chat=fake_chat, llm_gen_config=llm_gen_config)
    return agent_resp


def list_agents() -> List[XAgent]:
    """列出所有Agent的名称

    Returns:
        List[str]: Agent的名称列表
    """
    agents = []
    for agent_config_name in os.listdir(AGENT_DIR):
        # 加载agent
        agent_name = agent_config_name.replace(".json", "")
        try:
            agent = get_agent(name=agent_name)
            agents.append(agent)
        except Exception as e:
            logger.warning(f"fail to load agent {agent_name} with error:{e}")
            # raise e
    return agents


def clear_agent(name: str) -> None:
    """清空agent的历史对话记录

    Args:
        name (str): agent的名称
    """
    logger.info(f"clear memory of {name}")
    agent = get_agent(name=name)
    agent.clear_memory()


def delete_agent(name: str) -> None:
    logger.info(f"clear memory of {name}")
    agent = get_agent(name=name)
    agent.delete(AGENT_DIR)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/03/19 15:45:46
@Author  :   ChenHao
@Description  :   
@Contact :   jerrychen1990@gmail.com
'''
import copy
import os
import re
from typing import List, Type
from loguru import logger
from xagents.kb.kb_file import KnowledgeBaseFile
from xagents.kb.vector_store import XES, LocalVecStore, XVecStore, get_vecstore_cls
from xagents.loader.common import Chunk, ContentType
from xagents.model.api import get_embd_model, get_rerank_model
from xagents.kb.common import KnowledgeBaseInfo, RecalledChunk, chunk2document, get_chunk_dir, get_chunk_path, get_config_path, get_index_dir, get_kb_dir, get_origin_dir, get_status_path
from snippets import load, log_cost_time, dump, jload, jdump


class KnowledgeBase():

    def __init__(self, name: str, desc, embedding_config: dict, vecstore_config: dict | List[dict]):
        self.name = name
        self.desc = desc
        self.embedding_config = embedding_config
        self.vecstore_config = vecstore_config if isinstance(vecstore_config, list) else [vecstore_config]
        self.indexes = []
        self._build_dirs()
        self._save_config()

    @classmethod
    def from_config(cls, config: dict | str):
        if isinstance(config, str):
            # logger.info(f"{config=}")
            config = jload(config)
        logger.debug(f"{config=}")
        return cls(**config)

    def _get_config(self) -> dict:
        return dict(name=self.name, desc=self.desc, embedding_config=self.embedding_config, vecstore_config=self.vecstore_config)

    def get_info(self) -> KnowledgeBaseInfo:
        """返回知识库的信息

        Returns:
            dict: 知识库信息
        """
        file_num = len(list(self.list_kb_files()))
        return KnowledgeBaseInfo(**self._get_config(), file_num=file_num)

    def _build_dirs(self):
        self.kb_dir = get_kb_dir(self.name)
        self.origin_dir = get_origin_dir(self.name)

        self.chunk_dir = get_chunk_dir(self.name)
        self.config_path = get_config_path(self.name)
        self.status_path = get_status_path(self.name)

        os.makedirs(self.origin_dir, exist_ok=True)
        os.makedirs(self.chunk_dir, exist_ok=True)
        if not os.path.exists(self.config_path):
            jdump(dict(), self.status_path)

    def _save_config(self):
        dump(self._get_config(), self.config_path)

    def get_indexes(self) -> List[XVecStore]:
        """
        初始化向量存储、ES存储
        """
        if not self.indexes:
            for index_config in self.vecstore_config:
                config = copy.copy(index_config)
                vecstore_cls: Type[XVecStore] = get_vecstore_cls(config.pop("cls"))
                if vecstore_cls.is_local():
                    local_dir = get_index_dir(kb_name=self.name, index_name=f"{vecstore_cls.__name__}_index")
                    config.update(local_dir=local_dir)

                if vecstore_cls.need_embd():
                    embedding_model = get_embd_model(self.embedding_config)
                    config.update(embedding=embedding_model)
                if vecstore_cls == XES:
                    if "es_index" not in config:
                        config.update(es_index=f"{self.name}")

                logger.debug(f"creating index:{vecstore_cls} with config:{config}")
                index = vecstore_cls.from_config(config)
                self.indexes.append(index)
        return self.indexes

    def list_kb_files(self) -> List[KnowledgeBaseFile]:
        kb_files = []
        for file_name in os.listdir(self.origin_dir):
            kb_file = KnowledgeBaseFile(kb_name=self.name, file_name=file_name)
            kb_files.append(kb_file)
        return kb_files

    def remove_kb_file(self, kb_file: KnowledgeBaseFile):
        """
        删除文档
        """
        assert kb_file.kb_name == self.name
        # 从索引中删除
        self.remove_kb_file_from_index(kb_file=kb_file)
        # 删除kb文件
        kb_file.delete()

    def remove_kb_file_from_index(self, kb_file: KnowledgeBaseFile):
        ids = kb_file.get_chunk_ids()
        self.remove_chunks_from_index(ids)

    def remove_chunks_from_index(self, chunk_ids):
        for index in self.get_indexes():
            logger.info(f"deleting {len(chunk_ids)} chunks from index:{index}")
            index.delete(chunk_ids)
            if index.is_local():
                index.save()

    def add_chunks2index(self, chunks: List[Chunk], meta_info: dict, do_save=False, batch_size=16):
        indexes = self.get_indexes()
        chunks = [chunk for chunk in chunks if chunk.content]
        documents = [chunk2document(chunk, meta_info) for chunk in chunks]

        ids = [chunk.id for chunk in chunks]
        # 添加新的index
        for index in indexes:
            embd_model = index.embeddings
            if embd_model:
                if hasattr(embd_model, "batch_size"):
                    logger.debug(f"setting embedding model batch size to {batch_size}")
                    setattr(embd_model, "batch_size", batch_size)
                else:
                    logger.warning(f"{embd_model.__class__} has no attribute batch_size, set to default value {batch_size}")
            logger.debug(f"updating {len(ids)} documents to index:{index}")
            index.delete(ids=ids)
            index.add_documents(documents=documents, ids=ids)

            if index.is_local() and do_save:
                logger.info(f"saving to index dir:{index.local_dir}")
                index.save()

    def add_kb_file2index(self,  kb_file: KnowledgeBaseFile, do_save=False, batch_size=16):
        if not kb_file.is_cut:
            logger.warning(f"{kb_file.file_name} is not cut, please cut it first")
            return
        chunks = kb_file.list_chunks()
        kb_file.set_index_status(False)
        self.add_chunks2index(chunks=chunks, meta_info=dict(file_name=kb_file.file_name), do_save=do_save, batch_size=batch_size)
        kb_file.set_index_status(True)

    @log_cost_time(name="rebuild_index")
    def rebuild_index(self, batch_size: int):
        """
        重新构建向量知识库
        """
        indexes = self.get_indexes()
        kb_files = self.list_kb_files()
        for kb_file in kb_files:
            self.add_kb_file2index(kb_file=kb_file, do_save=False, batch_size=batch_size)
        for index in indexes:
            if isinstance(index, LocalVecStore):
                index.save()

    def delete(self):
        """删除知识库"""
        for index in self.get_indexes():
            index.delete_all()
        import shutil
        shutil.rmtree(path=self.kb_dir)

    @log_cost_time(name="kb_search")
    def search(self, query: str, top_k: int = 3, score_threshold: float = None,
               do_split_query=False, file_names: List[str] = None, rerank_config: dict = {},
               do_expand=False, expand_len: int = 500, forward_rate: float = 0.5) -> List[RecalledChunk]:
        """知识库检索

        Args:
            query (str): 待检索的query
            top_k (int, optional): 返回多少个chunk. Defaults to 3.
            score_threshold (float, optional): 召回的chunk相似度阈值. Defaults to None.
            do_split_query (bool, optional): 是否按照？切分query并分别召回. Defaults to False.
            file_names (List[str], optional): 按照名称过滤需要召回的片段所在的文件. Defaults to None.
            do_expand (bool, optional): 返回的chunk是否做上下文扩展. Defaults to False.
            expand_len (int, optional): 上下文扩展后的chunk字符长度（do_expand=True时生效）. Defaults to 500.
            forward_rate (float, optional): 上下文扩展时向下文扩展的比率（do_expand=True时生效）. Defaults to 0.5.

        Returns:
            List[RecalledChunk]: 相关的切片，按照score降序
        """
        recalled_chunks = []
        # 切分query
        queries = split_query(query, do_split_query)

        # 过滤条件
        _filter = dict()
        if file_names:
            _filter = dict(file_name=file_names)

        # 每个子query做检索
        indexes = self.get_indexes()
        for query in queries:

            for index in indexes:
                score_threshold = index.convert_score(score_threshold)
                logger.debug(f"searching {query} with vecstore_cls: {index.__class__.__name__}, {_filter=}, {top_k=}, {score_threshold=}")
                docs_with_score = index.similarity_search_with_score(query, k=top_k, score_threshold=score_threshold, filter=_filter)
                logger.debug(f"{docs_with_score=}")
                tmp_recalled_chunks = [RecalledChunk.from_document(d, score=index.convert_score(
                    s), query=query, index_cls=index.__class__.__name__) for d, s in docs_with_score]
                recalled_chunks.extend(tmp_recalled_chunks)

        # 去重，避免召回相同切片
        origin_recall_num = len(recalled_chunks)
        recalled_chunks = list(sorted(set(recalled_chunks), key=lambda x: x.score, reverse=True))
        logger.info(f"{len(recalled_chunks)} distinct recalled from {len(queries)} queries and {len(indexes)} indexes, {origin_recall_num-len(recalled_chunks)} duplicated")

        # 精排
        if rerank_config:
            recalled_chunks = rerank(recalled_chunks, rerank_config)
        recalled_chunks = recalled_chunks[:top_k]
        logger.info(f"get {len(recalled_chunks)} reranked chunks after sort")

        # 上下文扩展
        if do_expand:
            logger.info("expanding recalled chunks")
            for chunk in recalled_chunks:
                expand_chunk(chunk, self.name, expand_len, forward_rate)

        return recalled_chunks


def rerank(recalled_chunks: List[RecalledChunk], rerank_config: dict) -> List[RecalledChunk]:
    """重排序
    Args:
        recalled_chunks (List[RecalledChunk]): 待排序的切片

    Returns:
        List[RecalledChunk]: 排序后的切片
    """
    # logger.debug("reranking...")
    rerank_model = get_rerank_model(rerank_config)
    if rerank_model:
        logger.info("reranking chunks with rerank model")
        for chunk in recalled_chunks:
            similarity = rerank_model.cal_similarity(chunk.query, chunk.content)
            chunk.score = similarity

    recalled_chunks.sort(key=lambda x: x.score, reverse=True)
    return recalled_chunks


def split_query(query: str, do_split_query=False) -> List[str]:
    """切割query"""
    if do_split_query:
        rs = [e.strip() for e in re.split("\?|？", query) if e.strip()]
        logger.debug(f"split origin query into {len(rs)} queries")

        return rs
    else:
        # 不需要切分也转成list形式，方便后续统一处理
        return [query]


# 扩展上下文到给定的长度
# TODO 扩展时，避免重复的chunk
def expand_chunk(chunk: RecalledChunk, kb_name: str, expand_len: int, forward_rate=0.5) -> RecalledChunk:
    logger.debug(f"expanding chunk {chunk}")
    chunk_path = get_chunk_path(kb_name, chunk.file_name)
    chunk_dicts = load(chunk_path)
    chunk_idx = None

    for idx, ele in enumerate(chunk_dicts):
        if ele["id"] == chunk.id:
            chunk_idx = idx
            break

    if chunk_idx is None:
        logger.warning(f"chunk {chunk} not found in {chunk_path}")
        return chunk

    to_expand = expand_len - len(chunk.content)
    if to_expand <= 0:
        return chunk

    forward_len = int(to_expand * forward_rate)
    backward_len = to_expand - forward_len
    logger.debug(f"expand chunk with :{forward_len=}, {backward_len=}, origin_len:{len(chunk.content)}")
    backwards, forwards = [], []

    # 查找前面的chunk
    idx = chunk_idx-1
    while idx >= 0:
        if backward_len <= 0:
            break
        tmp_chunk = copy.copy(chunk_dicts[idx])
        if tmp_chunk["content_type"] == ContentType.TEXT:
            chunk_len = min(len(tmp_chunk["content"]), backward_len)
            tmp_chunk["content"] = tmp_chunk["content"][-chunk_len:]
            to_add: Chunk = Chunk(**tmp_chunk)
            backwards.append(to_add)
            backward_len -= chunk_len
        idx -= 1

    backwards.reverse()

    idx = chunk_idx + 1
    # logger.debug(f"{idx=}, {len(chunk_dicts)=}, {forward_len=}")
    while idx < len(chunk_dicts):
        if forward_len <= 0:
            break

        tmp_chunk = copy.copy(chunk_dicts[idx])
        if tmp_chunk["content_type"] == ContentType.TEXT:
            chunk_len = min(len(tmp_chunk["content"]), forward_len)
            tmp_chunk["content"] = tmp_chunk["content"][:chunk_len]
            to_add: Chunk = Chunk(**tmp_chunk)
            forwards.append(to_add)
            forward_len -= chunk_len
        idx += 1

    chunk.backwards = backwards
    chunk.forwards = forwards
    logger.debug(f"expand done with {len(backwards)} backward chunks and {len(forwards)} forward chunks, total_len:{chunk.total_len}")

    return chunk



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/04/24 10:36:09
@Author  :   ChenHao
@Description  : 知识库管理相关单测
@Contact :   jerrychen1990@gmail.com
'''

import os
from typing import List
from unittest import TestCase
from loguru import logger
from xagents.loader.common import Chunk, ContentType
from xagents.kb.common import DistanceStrategy, RecalledChunk
from xagents.config import DATA_DIR
from snippets import set_logger
from xagents.kb.api import add_chunks, create_knowledge_base, create_kb_file, delete_chunks, delete_kb_file, delete_knowledge_base, list_chunks, list_kb_files, search_knowledge_base, update_chunk

kb_name = "ut_kb"
kb_file_names = ["requirements.txt", "Xagents中间件.docx", "image_table.pdf"]
to_delete_kb_file_name = "Xagents中间件.docx"

def show_recall_chunks(recall_chunks:List[RecalledChunk]):
    for i, chunk in enumerate(recall_chunks):
        print(f"{i}: {chunk.to_plain_text()}")

# unit test
class TestKB(TestCase):

    @classmethod
    def setUpClass(cls):
        set_logger("dev", __name__)
        logger.info("start test KnowledgeBase")

    def test1_create_kb(self):
        vecstore_config = [dict(cls='XFAISS', distance_strategy=DistanceStrategy.MAX_INNER_PRODUCT)
                                                        # dict(cls="XES", es_url='http://localhost:9200')
                                                        ]
        kb  = create_knowledge_base(name=kb_name, desc="kb for unit test",
                                     embedding_config=dict(cls="ZhipuEmbedding"),
                                     vecstore_config = vecstore_config)

        kb_info = kb.get_info()
        logger.info(f"kb info: {kb_info}")

    def test2_add_kb_file(self):
        for kb_file_name in kb_file_names:
            kb_file_path = os.path.join(DATA_DIR, "kb_file", kb_file_name)
            kb_file = create_kb_file(kb_name=kb_name, file=kb_file_path)
            kb_file_info = kb_file.get_info()
            logger.info(f"kb file info: {kb_file_info}")
            
    def test3_list_kb_files(self):
        kb_files = list_kb_files(kb_name=kb_name)
        self.assertEqual(len(kb_files), len(kb_file_names))

        for kb_file in kb_files:
            kb_file_info = kb_file.get_info()
            logger.info(f"kb file info: {kb_file_info}")
            
    
    def test4_chunk_crud(self):
        chunks = list_chunks(kb_name=kb_name, file_name=to_delete_kb_file_name)
        logger.info(chunks[:2])
        logger.info(len(chunks))
        chunk = chunks[0]
        delete_chunks(kb_name=kb_name, file_name=to_delete_kb_file_name, chunk_ids=[chunk.id])
        chunks = list_chunks(kb_name=kb_name, file_name=to_delete_kb_file_name)
        logger.info(len(chunks))
        chunk = chunks[0]
        chunk.content = "updated_content"
        update_chunk(kb_name=kb_name, file_name=to_delete_kb_file_name, chunk=chunk)
        new_chunks = [Chunk.from_dict(dict(content="new_content", content_type=ContentType.TEXT))]
        add_chunks(kb_name=kb_name, file_name=to_delete_kb_file_name, chunks=new_chunks, idx=3)
        
        chunks = list_chunks(kb_name=kb_name, file_name=to_delete_kb_file_name)
        logger.info(len(chunks))
        for i, chunk in enumerate(chunks[:5]):
            logger.info(f"{i}: {chunk}")
    
            

            
    def test5_search_kb(self):
        query = "python-snippets"
        recall_chunks = search_knowledge_base(name=kb_name, query=query, top_k=5)
        show_recall_chunks(recall_chunks)
        
        query = "updated_content?new_content"
        recall_chunks = search_knowledge_base(name=kb_name, query=query, top_k=2)
        show_recall_chunks(recall_chunks)
        
        
        
    def test6_delete_kb_file(self):
        msg = delete_kb_file(kb_name=kb_name, file_name=to_delete_kb_file_name)
        logger.info(f"delete kb file: {msg}")
        
    def test7_delete_kb(self):
        msg = delete_knowledge_base(name=kb_name)
        logger.info(f"delete kb: {msg}")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/07 17:54:34
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''

import json
import itertools
from typing import Iterable, List, Optional

from abc import abstractmethod

from pydantic import BaseModel, Field

from agit.common import LLMResp, ToolCall
from xagents.config import DEFAULT_KB_PROMPT_TEMPLATE, DEFAULT_WEB_SEARCH_KB_PROMPT_TEMPLATE, \
    DEFAULT_WEB_SEARCH_PROMPT_TEMPLATE
from xagents.kb.common import RecalledChunk
from loguru import logger


class WebPage(BaseModel):
    url: str = Field(description="网页的URL")
    content: str = Field(description="网页的内容")


class WebSearchResp(BaseModel):
    summary: str = Field(description="搜索结果的摘要")
    pages: List[WebPage] | None = Field(description="搜索结果的网页列表")


class AgentResp(LLMResp):
    references: Optional[List[RecalledChunk]] = Field(description="召回的片段")
    web_search_result: Optional[WebSearchResp] = Field(description="联网搜索的结果", default=None)

    def to_stream(self):
        if self.references:
            for reference in self.references:
                reference = json.dumps(dict(reference=reference.model_dump()), ensure_ascii=False)
                yield reference+"\n"

        if self.tool_calls:
            for tool_call in self.tool_calls:
                tool_call = json.dumps(dict(tool_call=tool_call.model_dump()), ensure_ascii=False)
                yield tool_call+"\n"
        for ele in self.content:
            ele = json.dumps(dict(content_chunk=ele), ensure_ascii=False)
            yield ele+"\n"
        if self.usage:
            yield json.dumps(dict(usage=self.usage.model_dump()), ensure_ascii=False)+"\n"

    @classmethod
    def from_stream(cls, stream: Iterable):
        references = []
        tool_calls = []

        def json_stream_gen():
            for line in stream:
                if isinstance(line, bytes):
                    line = line.decode("utf-8")
                # logger.info(f"line: {line}")
                json_line = json.loads(line)
                yield json_line

        json_stream = json_stream_gen()

        for json_line in json_stream:
            if "reference" in json_line:
                references.append(RecalledChunk.model_validate(json_line["reference"]))
            elif "tool_call" in json_line:
                tool_calls.append(ToolCall.model_validate(json_line["tool_call"]))
            else:
                content_chunk = json_line
                break
        # logger.info(f"references: {references}")
        # logger.info(f"tool_calls: {tool_calls}")
        # logger.info(f"content_chunk: {content_chunk}")

        def content_gen():
            for json_line in itertools.chain([content_chunk], json_stream):
                yield json_line.get("content_chunk", "")
        return AgentResp(references=references, tool_calls=tool_calls, content=content_gen())


class AbstractAgent:

    def __init__(self, name) -> None:
        self.name = name

    @abstractmethod
    def chat(self, message: str, stream=True, do_remember=True) -> AgentResp:
        raise NotImplementedError

    @abstractmethod
    def remember(self, role: str, message: str):
        raise NotImplementedError


class KBConfig(BaseModel):
    name: str = Field(description="知识库名称")
    prompt_template: str = Field(description="应用知识库的提示词模板, 必须包含{context}和{question}两个字段",
                                 default=DEFAULT_KB_PROMPT_TEMPLATE)


class WebSearchConfig(BaseModel):
    name: str = Field(description="应用联网搜索的名称")
    prompt_search_template: str = Field(
        description="应用联网搜索的提示词模板, 必须包含{seach_info}和{question}两个字段",
        default=DEFAULT_WEB_SEARCH_PROMPT_TEMPLATE)
    prompt_search_kb_template: str = Field(
        description="应用知识库的提示词模板, 必须包含{search_info},{context}和{question}三个字段",
        default=DEFAULT_WEB_SEARCH_KB_PROMPT_TEMPLATE)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/11 16:40:25
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''


from abc import abstractmethod
import enum
import time
import uuid
from pathlib import Path
from typing import Iterable, List, Optional, Tuple

from pydantic import BaseModel, Field, ConfigDict, validator
from pandas import DataFrame
from numpy import ndarray
from xagents.config import *



# 切片类型


class ContentType(str, enum.Enum):
    TABLE = "TABLE"
    TITLE = "TITLE"
    TEXT = "TEXT"
    PARSED_TABLE = "PARSED_TABLE"
    IMAGE = "IMAGE"

# 切片

class BBox(BaseModel):
    x0: float
    y0: float
    x1: float
    y1: float
    @property
    def sort_key(self):
        return (self.y0, self.x0)
    
    def to_tuple(self):
        return (self.x0, self.y0, self.x1, self.y1)
    



class Chunk(BaseModel):
    id:str = Field(description="chunk的id", default_factory=lambda:str(uuid.uuid4()))
    content: str = Field(description="chunk的内容")
    content_type: ContentType = Field(description="chunk类型", default=ContentType.TEXT)
    search_content: Optional[str] = Field(description="用来检索的内容", default=None)
    page_idx: int = Field(description="chunk在文档中的页码,从1开始", default=1)
    
    def to_json(self):
        return self.model_dump(exclude_none=True)

    @classmethod
    def from_dict(cls, data:dict) -> "Chunk":
        content_type = data["content_type"]
        cls:type[BaseModel] =_content_type2cls[content_type]
        return cls.model_validate(data)
        
class TextChunk(Chunk):
    content_type:ContentType=ContentType.TEXT
    
class ImageChunk(Chunk):
    content_type:ContentType=ContentType.IMAGE
    content: Optional[str] = Field(description="image的文字描述", default=None)
    url:Optional[str] = Field(description="图片的url", default=None)
    image_name:str = Field(description="图片的文件名")

class TableChunk(Chunk):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    content_type:ContentType=ContentType.TABLE
    data: Optional[ndarray|DataFrame] = Field(description="表格的数据, 二维数组或pandas的dataframe", default=None)
    content: str = Field(description="表格的文字描述, markdown格式", default=None)
    
    def to_json(self):
        return self.model_dump(exclude_none=True, exclude={"data"})

    @validator('data', pre=True, always=True)
    def set_content_based_on_data(cls, v, values, **kwargs):
        # logger.info(f"{values=}")
        # Check if description is already set
        if values.get("content") is None:
            content = data2markdown(v)
            values["content"] = content
            return v
        return v    


_content_type2cls={
    ContentType.TABLE: TableChunk,
    ContentType.TEXT: TextChunk,
    ContentType.IMAGE: ImageChunk,
}

class AbstractLoader:
    def __init__(self, **kwargs) -> None:
        pass
    @abstractmethod
    def load(self, file_path: str,  start_page:int=0, end_page:int=None, store_image=True, ocr=False, **kwargs) -> List[Chunk]:
        raise NotImplementedError


def merge_chunks(chunks: Iterable[Chunk], joiner="\n") -> Iterable[Chunk]:
    acc = None
    cur_page_idx=1
    for chunk in chunks:
        cur_page_idx = chunk.page_idx
        if chunk.content_type == ContentType.TEXT:
            acc = joiner.join([acc, chunk.content]) if acc else chunk.content
        else:
            if acc:
                yield TextChunk(content=acc, page_idx=cur_page_idx)
                acc = None
            yield chunk
    if acc:
        yield TextChunk(content=acc, page_idx=cur_page_idx)
            
    

def get_image_name_path(kf_file_name:str, page_idx:int, image_idx:int)->Tuple[str, str]:
    image_name =f"{Path(kf_file_name).stem}-page{page_idx}-image{image_idx}-time{time.time()}.png"
    file_path = os.path.join(TEMP_DIR, "extracted_images", image_name)
    os.makedirs(os.path.dirname(file_path), exist_ok=True)
    return image_name, file_path

def data2markdown(data:ndarray|DataFrame)->str:
    if isinstance(data, ndarray):
        data = DataFrame(data)
    data.fillna("", inplace=True)
    markdown = data.to_markdown(index=False)
    return markdown



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/02/20 15:47:53
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''


from typing import List


import requests
from snippets import *
from xagents.model.common import LLM, EMBD
from agit.llm import LLM_TYPE, call_llm, LLMResp
from agit.embd import EMBD_TYPE, call_embedding

from loguru import logger


def call_local_embedding(url: str, contents: List[str], norm=True) -> List[List[float]]:
    # embeddings = []
    # url = "http://36.103.177.140:8001/v2/embeddings"
    resp = requests.post(url=url, json=dict(texts=contents, norm=norm))
    resp.raise_for_status()
    return resp.json()["data"]['embeddings']


class LocalEmbedding(EMBD):

    def __init__(self,  url="http://hz-model.bigmodel.cn/embedding-models/v2/embeddings", batch_size=16, norm=True, dim=1024):
        self.batch_size = batch_size
        self.norm = norm
        self.url = url
        self.dim=dim

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        logger.info(f"embedding {len(texts)} texts with {self.batch_size=}")
        embeddings = call_embedding(text=texts,model_or_url=self.url, embd_type=EMBD_TYPE.LOCAL,
                                    norm=self.norm, batch_size=self.batch_size)
        
        return embeddings

    def embed_query(self, text: str) -> List[float]:
        embedding = call_embedding(text=text,model_or_url=self.url, embd_type=EMBD_TYPE.LOCAL,
                                    norm=self.norm, batch_size=self.batch_size)
        return embedding

    def get_dim(self) -> int:
        return self.dim


class TGI_GLM(LLM):
    """
    本地调用LLM
    """

    @classmethod
    def list_versions(cls):
        return [
            "v1.0.0",
        ]

    def __init__(self, url:str, name=None, version="v1.0.0"):
        super().__init__(name=name, version=version)
        self.url = url

    def generate(self, prompt:str, history=[], system:str=None, stream=True, **kwargs)->LLMResp:
        messages = history + [{"role": "user", "content": prompt}]
        resp = call_llm(messages=messages, url=self.url, llm_type=LLM_TYPE.TGI, system=system,
                        stream=stream, **kwargs)

        return resp



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/04/16 10:43:40
@Author  :   ChenHao
@Description  : job模块的测试脚本
@Contact :   jerrychen1990@gmail.com
'''


from concurrent.futures import ProcessPoolExecutor
 
import time
from unittest import TestCase
from loguru import logger
from snippets import set_logger
from xagents.service.job import Job, JobStatus, get_job_info, init_job_db, stop_job, list_job_infos
from xagents.config import *


def long_time_job(num):
    # print(num)
    logger.info(f"job with {num} seconds started")
    for i in range(num):
        time.sleep(1)
        # print(f"{i+1} seconds passed")
        logger.info(f"{i+1} seconds passed")
    logger.info(f"job with {num} seconds finished")
    return num**2


# unit test
class TestJob(TestCase):

    @classmethod
    def setUpClass(cls):
        set_logger("dev", __name__)
        init_job_db()
        logger.info("start test job")
        

    def test_list_job_infos(self):
        job_infos = list_job_infos(status=[JobStatus.SUCCESS, JobStatus.FAILED])
        self.assertTrue(len(job_infos)>0, "The list is empty")
        for job_info in job_infos:
            logger.info(f"{job_info=}")        

        
    def test_job(self):
        job_ids =[]
        pool = ProcessPoolExecutor(max_workers=2)
        for t in [5, 10]:
            job:Job = Job.create_job(func=long_time_job, kwargs=dict(num=t)) 
            job_ids.append(job.job_id)
            pool.submit(job.run)
                
        for _ in range(4):
            job_infos = list_job_infos(status=JobStatus.RUNNING)
            for job_info in job_infos:
                logger.info(f"{job_info=}")
            logger.info("waiting 2 seconds...")
            time.sleep(2)
            
        for job_id in job_ids:
            stop_job(job_id=job_id)
            
        # 检测job状态
        EXPECTS = [JobStatus.SUCCESS, JobStatus.STOP]
        for idx, job_id in enumerate(job_ids):
            job_info = get_job_info(job_id)
            logger.info(f"{job_info=}")
            
            self.assertEqual(job_info.status, EXPECTS[idx])
        pool.shutdown()
            
            




#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/27 19:58:16
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''


# from xagents.model.core import LLM
# from agit.backend.openai_bk import call_llm_api

# from loguru import logger

# class GPT(LLM):
#     def __init__(self, name: str, version: str, proxy="zhipu"):
#         super().__init__(name, version)
#         self.proxy = proxy

#     @classmethod
#     def list_versions(cls):
#         return ["gpt-4"]
#         # models = list_models(keyword="gpt", proxy=self.proxy)
#         # model_names = sorted([e.id for e in models])
#         # return model_names

#     def generate(self, prompt, history=[], system=None, stream=True, temperature=0.01, **kwargs):
#         resp = call_llm_api(prompt=prompt, history=history, model=self.version, proxy=self.proxy,
#                             temperature=temperature, system=system, stream=stream, logger=logger, **kwargs)
#         return resp


# if __name__ == "__main__":
#     gpt = GPT(name="gpt", version="gpt-4")
#     # print(gpt.list_versions())
#     resp = gpt.generate(prompt="你是谁", system="把我的话翻译成英语")
#     for item in resp:
#         print(item, end="")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/01/04 14:04:02
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''
import sys
from xagents.tool.core import BaseTool
from agit.common import Parameter


def exec_python(expression: str) -> dict:
    try:
        rs = eval(expression)
    except Exception as e:
        rs = "执行失败"
    return dict(result=rs)


def mock_travel_search(departure: str, destination: str, date: str):
    return dict(result="查询成功，有如下车次:A00001, A00002, A00003")


calculator = BaseTool(name="计算器", description="根据提供的数学表达式，用python解释器来执行，得到计算结果,计算结果以json格式来返回,json包含一个字段，名字为result",
                      parameters=[Parameter(name="expression", type="string", description="数学表达式，可以通过python来执行的", required=True)],
                      callable=exec_python)

travel_searcher = BaseTool(name="车次查询", description="根据用户提供的信息，查询对应的车次",
                           parameters=[Parameter(name="departure", description="出发城市或车站", type="string", required=True),
                                       Parameter(name="destination", description="目的地城市或车站", type="string", required=True),
                                       Parameter(name="date", description="要查询的车次日期", type="string", required=True)],
                           callable=mock_travel_search)


if __name__ == "__main__":
    rs = calculator.execute("(351345-54351)/54351")
    print(rs)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/01/04 18:26:54
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''

from typing import Any, List

from xagents.tool.common_tool import calculator, travel_searcher
from xagents.tool.web_search import web_search_sogou
from xagents.tool.core import BaseTool, ToolCall
from loguru import logger


_ALL_TOOLS = [calculator, travel_searcher, web_search_sogou]
_TOOL_MAP = {e.name:e for e in _ALL_TOOLS}


def invoke_tool_call(tool_call:ToolCall)->Any:
    tool = get_tool(tool_call.name)
    resp = tool.execute(**tool_call.parameters)
    tool_call.resp = resp
    return resp
    

def get_tool(tool_name:str)->BaseTool:
    if tool_name not in _TOOL_MAP:
        raise ValueError(f"Unknown tool: {tool_name}")
    tool:BaseTool = _TOOL_MAP[tool_name]
    return tool
    
def list_tools() -> List[BaseTool]:
    return _ALL_TOOLS
    
def get_tools(tools_config:List[dict]) -> List[BaseTool]:
    tools = []
    for config in tools_config:
        try:
            tool_name = config.get("name")
            tool = get_tool(tool_name)
            tools.append(tool)   
        except Exception as e:
            logger.warning(f"Failed to load tool: {tool_name}")
    return tools


if __name__ == "__main__":
    tool_call = ToolCall(name="calculator", parameters={"expression": "1+1"})
    resp = invoke_tool_call(tool_call)
    print(resp)    
    



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/03/11 14:22:25
@Author  :   ChenHao
@Description  :   
@Contact :   jerrychen1990@gmail.com
'''

import sys

from setuptools import find_packages, setup

from snippets.utils import get_latest_version, get_next_version, read2list

def get_install_req():
    req = read2list("requirements.txt")
    return req



if __name__ == "__main__":
    name = "xagent"
    if len(sys.argv) >= 4 and sys.argv[-1].startswith("v"):
        version = sys.argv.pop(-1)
    else:
        latest_version = get_latest_version(name)
        version = get_next_version(latest_version)
    print(f"version: {version}")
    install_req = get_install_req()
    print(f"install_req: {install_req}")
    setup(
        name=name,
        version=version,
        install_requires=install_req,
        packages=find_packages(exclude=['tests*']),
        package_dir={"": "."},
        package_data={},
        url='https://dev.aminer.cn/solution-center-algorithm/XAgents',
        license='MIT',
        author='Chen Hao',
        author_email='jerrychen1990@gmail.com',
        zip_safe=True,
        description='xagent',
        long_description="xagent"
    )



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/04/11 14:22:06
@Author  :   ChenHao
@Description  :   智谱API model的单测
@Contact :   jerrychen1990@gmail.com
'''
import json
import os
from unittest import TestCase
from xagents.model.zhipu import API_GLM
from xagents.model.common import EMBD, LLMResp
from xagents.model import get_embd_model, get_llm_model
from xagents.tool import travel_searcher
from xagents.tool.api import invoke_tool_call
from loguru import logger
from snippets import set_logger


# unit test
class TestEMBD(TestCase):

    @classmethod
    def setUpClass(cls):
        set_logger("dev", __name__)
        logger.info("start test embd")
        cls.zhipu_api_embedding: EMBD = get_embd_model(dict(cls="ZhipuEmbedding", api_key=os.environ["ZHIPU_API_KEY"], batch_size=4))

    def test_zhipu_api(self):
        # set_logger("dev", "")
        texts = ["你好", "hello"]
        embds = self.zhipu_api_embedding.embed_documents(texts)
        logger.info(len(embds))
        self.assertEqual(len(embds), 2)
        import numpy as np
        logger.info(np.linalg.norm(embds[0]))
        self.assertAlmostEqual(np.linalg.norm(embds[0]), 1.0)

        print(embds[0][:4])
        embd = self.zhipu_api_embedding.embed_query(text=texts[0])
        # print(embd[:4])
        # self.assertListEqual(embds[0], embd)


class TestGLM(TestCase):

    @classmethod
    def setUpClass(cls):
        set_logger("dev", __name__)
        logger.info("start llm")
        cls.zhipu_api_llm_model: API_GLM = get_llm_model(
            dict(cls="GLM", api_key=os.environ["ZHIPU_API_KEY"], name="glm-3-turbo", version="glm-3-turbo"))

    def test_zhipu_api(self):
        # set_logger("dev", "")
        prompt = "你好呀，你是谁"
        _system = "请用英语回答我的问题，你的名字叫XAgent"

        # 测试zhipu api
        resp: LLMResp = self.zhipu_api_llm_model.generate(prompt=prompt, system=_system, temperature=0.7,
                                                          top_p=0.95, max_tokens=100, log_level="INFO", stream=False)

        logger.info(json.dumps(resp.model_dump(), ensure_ascii=False, indent=4))
        self.assertIsNotNone(resp.content)
        self.assertIsNotNone(resp.usage)

        # 流式
        resp: LLMResp = self.zhipu_api_llm_model.generate(prompt=prompt, system=_system, temperature=0.7,
                                                          top_p=0.95, max_tokens=100, log_level="INFO", stream=True)
        for chunk in resp.content:
            logger.info(chunk)
        logger.info(json.dumps(resp.model_dump(exclude={"content"}), ensure_ascii=False, indent=4))
        self.assertIsNotNone(resp.usage)

    def test_zhipu_api_tool_call(self):

        prompt = "你能帮我查询2024年1月1日从北京南站到上海的火车票吗？"
        _system = None
        tools = [travel_searcher]
        # 测试zhipu api
        resp: LLMResp = self.zhipu_api_llm_model.generate(prompt=prompt, system=_system, tools=tools,
                                                          temperature=0.7, top_p=0.95, max_tokens=100, log_level="INFO", stream=False)
        logger.info(json.dumps(resp.model_dump(), ensure_ascii=False, indent=4))
        self.assertIsNotNone(resp.tool_calls)
        self.assertIsNotNone(resp.usage)
        # 调用工具
        logger.info("invoking tools")
        for tool_call in resp.tool_calls:
            tool_resp = invoke_tool_call(tool_call)

            tool_call.resp = tool_resp
            resp: LLMResp = self.zhipu_api_llm_model.observe(prompt=prompt, tool_call=tool_call, tools=tools, stream=False)

            logger.info(json.dumps(resp.model_dump(), ensure_ascii=False, indent=4))
            self.assertIsNotNone(resp.content)
            self.assertIsNotNone(resp.usage)

        # ## 流式
        resp = self.zhipu_api_llm_model.generate(prompt=prompt, system=_system, tools=tools, temperature=0.7,
                                                 top_p=0.95, max_tokens=100, log_level="INFO", stream=False)
        logger.info(json.dumps(resp.model_dump(exclude={"content"}), ensure_ascii=False, indent=4))

        self.assertIsNotNone(resp.tool_calls)
        self.assertIsNotNone(resp.usage)

        # 调用工具
        logger.info("invoking tools")
        for tool_call in resp.tool_calls:
            tool_resp = invoke_tool_call(tool_call)

            tool_call.resp = tool_resp
            resp: LLMResp = self.zhipu_api_llm_model.observe(prompt=prompt, tool_call=tool_call, tools=tools, stream=True)
            for chunk in resp.content:
                logger.info(chunk)
            logger.info(json.dumps(resp.model_dump(exclude={"content"}), ensure_ascii=False, indent=4))
            self.assertIsNotNone(resp.usage)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/04/11 14:22:06
@Author  :   ChenHao
@Description  :   智谱API model的单测
@Contact :   jerrychen1990@gmail.com
'''
import json
from unittest import TestCase
from xagents.model.api import get_llm_model
from xagents.model.common import EMBD, LLM, LLMResp
from xagents.model import get_embd_model
from loguru import logger
from snippets import set_logger
from agit.utils import cal_vec_similarity


# unit test
class TestEMBD(TestCase):

    @classmethod
    def setUpClass(cls):
        set_logger("dev", __name__)
        logger.info("start test embd")
        cls.local_embd:EMBD= get_embd_model(dict(cls="LocalEmbedding", batch_size=4, norm=False))

    def test_local_embd(self):
        # set_logger("dev", "")
        texts = ["你好", "hello"]
        embds = self.local_embd.embed_documents(texts)
        logger.info(len(embds))
        self.assertEqual(len(embds), 2)
        self.assertEqual(len(embds[0]), self.local_embd.get_dim())
        import numpy as np
        logger.info(np.linalg.norm(embds[0]))
        self.assertNotEqual(np.linalg.norm(embds[0]), 1.0)

        print(embds[0][:4])
        embd = self.local_embd.embed_query(text=texts[0])
        print(embd[:4])
        
        cosine_simi = cal_vec_similarity(embds[0], embd)
        logger.info(f"cosine similarity: {cosine_simi}")
        self.assertAlmostEquals(cosine_simi, 1.)    
            
class TestTGI(TestCase):

    @classmethod
    def setUpClass(cls):
        set_logger("dev", __name__)
        logger.info("start llm")
        cls.tgi_llm_model:LLM = get_llm_model(dict(cls="TGI_GLM", url="http://hz-model.bigmodel.cn/wind-130b"))

    def test_tgi_llm(self):
        # set_logger("dev", "")
        prompt = "你好呀，你是谁"
        _system = "请用英语回答我的问题，你的名字叫XAgent"
        
        # 测试zhipu api
        resp:LLMResp = self.tgi_llm_model.generate(prompt=prompt, system=_system, temperature=0.7, top_p=0.95, max_tokens=100, log_level="INFO", stream=False)        

        logger.info(json.dumps(resp.model_dump(), ensure_ascii=False, indent=4))
        self.assertIsNotNone(resp.content)        


        ## 流式
        resp:LLMResp = self.tgi_llm_model.generate(prompt=prompt, system=_system, temperature=0.7, top_p=0.95, max_tokens=100, log_level="INFO", stream=True)        
        for chunk in resp.content:
            logger.info(chunk)
        logger.info(json.dumps(resp.model_dump(exclude={"content"}), ensure_ascii=False, indent=4))

  


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/11 16:42:25
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''

import numpy as np
from loguru import logger
from typing import Iterable, List, Tuple


from xagents.loader.utils import upload_to_minio, image2text
from xagents.config import *
from xagents.loader.common import BBox, Chunk, TableChunk, TextChunk, ImageChunk, get_image_name_path, merge_chunks, AbstractLoader
from pdfplumber.page import Page


def analyse_page(idx: int, page: Page, file_name: str, upload_image: bool, ocr: bool) -> Iterable[Chunk]:

    chunks: List[Tuple[Chunk, BBox]] = []  # List to store all content with their position

    # 抽取文本
    for line in page.extract_text_lines(x_tolerance=3, y_tolerance=3):
        chunks.append((TextChunk(content=line['text'], page_idx=idx), BBox(x0=line["x0"], y0=line["top"], x1=line["x1"], y1=line["bottom"])))

    tables = page.extract_tables()
    for table in tables:
        bbox = BBox(x0=0, x1=0, y0=0, y1=0)
        # logger.info(f"table: {table}")

        table = np.array(table)
        chunks.append((TableChunk(data=table, page_idx=idx), bbox))

    # Extract images and add to content list
    page_image = page.to_image().original
    for i, image in enumerate(page.images):
        bbox = BBox(x0=image['x0'], y0=image['top'], x1=image['x1'], y1=image['bottom'])
        # pil_image = Image.open(page_image.annotated)
        image = page_image.crop(bbox.to_tuple())

        image_name, image_path = get_image_name_path(file_name, idx, i)
        if upload_image:
            os.makedirs(os.path.dirname(image_path), exist_ok=True)
            image.save(image_path, format="PNG")
            url = upload_to_minio(image_path)
        else:
            url = None
        if ocr:
            content = image2text(image)
        else:
            content = None
        chunks.append((ImageChunk(url=url, content=content, image_name=image_name, page_idx=idx), bbox))

    chunks.sort(key=lambda x: x[1].sort_key)  # Sort by top y coordinate, then by left x coordinate
    yield from (x[0] for x in chunks)


class PDFLoader(AbstractLoader):
    def load(self, file_path: str, start_page=1, end_page=None, upload_image=True, ocr=False) -> Iterable[Chunk]:
        import pdfplumber

        logger.info(f"loading pdf file {file_path} from {start_page=} to {end_page=}, {upload_image=}, {ocr=}")
        pages = pdfplumber.open(file_path)
        for idx, page in enumerate(pages.pages, start=1):
            if idx < start_page:
                continue
            if end_page and idx > end_page:
                break
            chunks = analyse_page(idx, page, os.path.basename(file_path), upload_image=upload_image, ocr=ocr)
            yield from merge_chunks(chunks=chunks)


if __name__ == "__main__":
    from xagents.config import XAGENT_HOME
    import sys
    import pdfplumber
    logger.add(sys.stdout)
    doc_path = os.path.join(XAGENT_HOME, "data/kb_file/image_table.pdf")
    pages = pdfplumber.open(doc_path)
    pages = list(pages.pages)
    contents = analyse_page(pages[4])
    for content in contents:

        print(content)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/04/12 15:00:12
@Author  :   ChenHao
@Description  :  doc/docx文档解析
@Contact :   jerrychen1990@gmail.com
'''



from typing import Iterable

import docx.table


from xagents.loader.utils import upload_to_minio, image2text
from xagents.config import *
from xagents.loader.common import Chunk, AbstractLoader, ImageChunk, TextChunk, TableChunk, get_image_name_path, merge_chunks

import docx
import pandas as pd

from PIL import Image
from docx.text.paragraph import Paragraph
import xml.etree.ElementTree as ET
from docx.document import Document as DOC
from docx.oxml.table import CT_Tbl
from docx.oxml.text.paragraph import CT_P
from docx.table import _Cell
from xml.etree import ElementTree
from io import StringIO
import io
import csv

def iter_block_items(parent):
    """
    Yield each paragraph and table child within *parent*, in document order.
    Each returned value is an instance of either Table or Paragraph. *parent*
    would most commonly be a reference to a main Document object, but
    also works for a _Cell object, which itself can contain paragraphs and tables.
    """
    if isinstance(parent, DOC):
        parent_elm = parent.element.body
    elif isinstance(parent, _Cell):
        parent_elm = parent._tc
    else:
        raise ValueError("something's not right")

    for child in parent_elm.iterchildren():
        if isinstance(child, CT_P):
            yield Paragraph(child, parent)
        elif isinstance(child, CT_Tbl):
            yield docx.table.Table(child, parent)

    
#This function extracts the table from the document object as a dataframe
def read_docx_tables(document, tab_id=None, **kwargs):
    """
    parse table(s) from a Word Document (.docx) into Pandas DataFrame(s)

    Parameters:
        filename:   file name of a Word Document

        tab_id:     parse a single table with the index: [tab_id] (counting from 0).
                    When [None] - return a list of DataFrames (parse all tables)

        kwargs:     arguments to pass to `pd.read_csv()` function

    Return: a single DataFrame if tab_id != None or a list of DataFrames otherwise
    """
    def read_docx_tab(tab, **kwargs):
        vf = io.StringIO()
        writer = csv.writer(vf)
        for row in tab.rows:
            writer.writerow(cell.text for cell in row.cells)
        vf.seek(0)
        return pd.read_csv(vf, **kwargs)

#    doc = Document(filename)
    if tab_id is None:
        return [read_docx_tab(tab, **kwargs) for tab in document.tables]
    else:
        try:
            return read_docx_tab(document.tables[tab_id], **kwargs)
        except IndexError:
            print('Error: specified [tab_id]: {}  does not exist.'.format(tab_id))
            raise

def analyse_page(document:DOC, file_name:str, upload_image:bool, ocr:bool)->Iterable[Chunk]:

    page_index = 1
    table_idx=0
    image_idx = 0

    for block in iter_block_items(document):
        if 'text' in str(block):
        
            content = str(block.text)
            yield TextChunk(content=content, page_idx=page_index)

            for run in block.runs:
                if 'pageBreakBefore' in run.element.xml or '\f' in run.text:
                    page_index += 1
                xmlstr = str(run.element.xml)
                my_namespaces = dict([node for _, node in ElementTree.iterparse(StringIO(xmlstr), events=['start-ns'])])
                root = ET.fromstring(xmlstr) 
                #Check if pic is there in the xml of the element. If yes, then extract the image data
                if 'pic:pic' in xmlstr:
                    for pic in root.findall('.//pic:pic', my_namespaces):
                        blip_elem = pic.find("pic:blipFill/a:blip", my_namespaces)
                        embed_attr = blip_elem.get("{http://schemas.openxmlformats.org/officeDocument/2006/relationships}embed")
                        document_part = document.part
                        image_part = document_part.related_parts[embed_attr]
                        image_name, image_path =get_image_name_path(file_name, 0, image_idx)    

                        if upload_image:
                            with open(image_path, "wb") as f:
                                f.write(image_part._blob)
                
                            url = upload_to_minio(image_path)
                        else:
                            url=None
                        content = image2text(Image.open(io.BytesIO(image_part._blob))) if ocr else None
                        # image_base64 = base64.b64encode(image_part._blob)
                        # image_base64 = image_base64.decode()        
                        yield ImageChunk(url=url, content=content,page_idx=page_index, image_name=image_name)
                    image_idx = image_idx + 1
                
        elif 'table' in str(block):
            content = str(block)
            df:pd.DataFrame = read_docx_tables(document=document,tab_id=table_idx)
            yield TableChunk(page_idx=page_index, data=df)
            table_idx+=1

            
class DOCXLoader(AbstractLoader):
    def __init__(self, max_page:int=None, **kwargs):
        """构建pdf加载器

        Args:
            max_page (int, optional): 最大页数. Defaults to None：不限定页数
            extract_images (bool, optional): 是否使用ocr抽取其中的图片. Defaults to False.
        """
        super().__init__(**kwargs)
        self.max_page = max_page


    def load(self, file_path: str, start_page=1, end_page=None, upload_image=True, ocr=False, **kwargs) -> Iterable[Chunk]:
        #TODO 缺少页码控制
        doc = docx.Document(file_path)
        chunks = analyse_page(doc, file_name=os.path.basename(file_path), upload_image=upload_image, ocr=ocr)
        chunks = merge_chunks(chunks)
        yield from chunks


if __name__ == "__main__":
    doc = docx.Document("data/kb_file/Xagents中间件.docx")
    chunks = analyse_page(doc)
    for chunk in chunks:
        print(chunk)